"""
Module for main signal processing operations.

.. autosummary::
    :toctree: generated/

    to_array
    calculate_energy
"""

import numpy as np

__all__ = [
    "SAMPLE_WIDTH_TO_DTYPE",
    "to_array",
    "calculate_energy",
]

SAMPLE_WIDTH_TO_DTYPE = {1: np.int8, 2: np.int16, 4: np.int32}
EPSILON = 1e-10


def _get_numpy_dtype(sample_width):
    """
    Helper function to convert a sample width to the corresponding NumPy data
    type.

    Parameters
    ----------
    sample_width : int
        The width of the sample in bytes. Accepted values are 1, 2, or 4.

    Returns
    -------
    numpy.dtype
        The corresponding NumPy data type for the specified sample width.

    Raises
    ------
    ValueError
        If `sample_width` is not one of the accepted values (1, 2, or 4).
    """

    dtype = SAMPLE_WIDTH_TO_DTYPE.get(sample_width)
    if dtype is None:
        err_msg = "'sample_width' must be 1, 2 or 4, given: {}"
        raise ValueError(err_msg.format(sample_width))
    return dtype


def to_array(data, sample_width, channels):
    """
    Convert raw audio data into a NumPy array.

    This function transforms raw audio data, specified by sample width and
    number of channels, into a 2-D NumPy array of `numpy.float64` data type.
    The array will be arranged by channels and samples.

    Parameters
    ----------
    data : bytes
        The raw audio data.
    sample_width : int
        The sample width (in bytes) of each audio sample.
    channels : int
        The number of audio channels.

    Returns
    -------
    numpy.ndarray
        A 2-D NumPy array representing the audio data. The shape of the array
        will be (number of channels, number of samples), with data type
        `numpy.float64`.

    Raises
    ------
    ValueError
        If `sample_width` is not an accepted value for conversion by the helper
        function `_get_numpy_dtype`.
    """

    dtype = _get_numpy_dtype(sample_width)
    array = np.frombuffer(data, dtype=dtype).astype(np.float64)
    return array.reshape(channels, -1, order="F")


def calculate_energy(x, agg_fn=None):
    """Calculate the energy of audio data.

    The energy is calculated as:

    .. math::
       \\text{energy} = 20 \\log(\\sqrt({1}/{N} \\sum_{i=1}^{N} {a_i}^2))  % # noqa: W605

    where `a_i` is the i-th audio sample and `N` is the total number of samples
    in `x`.

    Parameters
    ----------
    x : array
        Array of audio data, which may contain multiple channels.
    agg_fn : callable, optional
        Aggregation function to use for multi-channel data. If None, the energy
        will be computed and returned separately for each channel.

    Returns
    -------
    float or numpy.ndarray
        The energy of the audio signal. If `x` is multichannel and `agg_fn` is
        None, this will be an array of energies, one per channel.
    """

    x = np.array(x).astype(np.float64)
    energy_sqrt = np.sqrt(np.mean(x**2, axis=-1))
    energy_sqrt = np.clip(energy_sqrt, a_min=EPSILON, a_max=None)
    energy = 20 * np.log10(energy_sqrt)
    if agg_fn is not None:
        energy = agg_fn(energy)
    return energy



import math
import os
from pathlib import Path
from random import random
from shutil import which
from tempfile import TemporaryDirectory
from unittest import mock
from unittest.mock import patch

import numpy as np
import pytest

from auditok import (
    AudioParameterError,
    AudioRegion,
    load,
    make_silence,
    split,
    split_and_join_with_silence,
)
from auditok.audio import (
    _duration_to_nb_windows,
    _make_audio_region,
    _read_chunks_online,
    _read_offline,
)
from auditok.io import get_audio_source
from auditok.signal import to_array
from auditok.util import AudioReader

requires_ffmpeg = pytest.mark.skipif(
    which("ffmpeg") is None, reason="ffmpeg not available"
)

mock._magics.add("__round__")


def _make_random_length_regions(
    byte_seq, sampling_rate, sample_width, channels
):
    regions = []
    for b in byte_seq:
        duration = round(random() * 10, 6)
        data = b * int(duration * sampling_rate) * sample_width * channels
        region = AudioRegion(data, sampling_rate, sample_width, channels)
        regions.append(region)
    return regions


@pytest.mark.parametrize(
    "skip, max_read, channels",
    [
        (0, -1, 1),  # no_skip_read_all
        (0, -1, 2),  # no_skip_read_all_stereo
        (2, -1, 1),  # skip_2_read_all
        (2, None, 1),  # skip_2_read_all_None
        (2, 3, 1),  # skip_2_read_3
        (2, 3.5, 2),  # skip_2_read_3_5_stereo
        (2.4, 3.5, 2),  # skip_2_4_read_3_5_stereo
    ],
    ids=[
        "no_skip_read_all",
        "no_skip_read_all_stereo",
        "skip_2_read_all",
        "skip_2_read_all_None",
        "skip_2_read_3",
        "skip_2_read_3_5_stereo",
        "skip_2_4_read_3_5_stereo",
    ],
)
def test_load(skip, max_read, channels):
    sampling_rate = 10
    sample_width = 2
    filename = "tests/data/test_split_10HZ_{}.raw"
    filename = filename.format("mono" if channels == 1 else "stereo")
    region = load(
        filename,
        skip=skip,
        max_read=max_read,
        sr=sampling_rate,
        sw=sample_width,
        ch=channels,
    )
    with open(filename, "rb") as fp:
        fp.read(round(skip * sampling_rate * sample_width * channels))
        if max_read is None or max_read < 0:
            to_read = -1
        else:
            to_read = round(max_read * sampling_rate * sample_width * channels)
        expected = fp.read(to_read)
    assert bytes(region) == expected


@pytest.mark.parametrize(
    "duration, sampling_rate, sample_width, channels",
    [
        (1.05, 16000, 1, 1),  # mono_16K_1byte
        (1.5, 16000, 2, 1),  # mono_16K_2byte
        (1.0001, 44100, 2, 2),  # stereo_44100_2byte
        (1.000005, 48000, 2, 3),  # 3channel_48K_2byte
        (1.0001, 48000, 4, 4),  # 4channel_48K_4byte
        (0, 48000, 4, 4),  # 4channel_4K_4byte_0sec
    ],
    ids=[
        "mono_16K_1byte",
        "mono_16K_2byte",
        "stereo_44100_2byte",
        "3channel_48000_2byte",
        "4channel_48K_4byte",
        "4channel_4K_4byte_0sec",
    ],
)
def test_make_silence(duration, sampling_rate, sample_width, channels):
    silence = make_silence(duration, sampling_rate, sample_width, channels)
    size = round(duration * sampling_rate) * sample_width * channels
    expected_data = b"\0" * size
    expected_duration = size / (sampling_rate * sample_width * channels)
    assert silence.duration == expected_duration
    assert silence.data == expected_data


@pytest.mark.parametrize(
    "duration",
    [
        (0,),  # zero_second
        (1,),  # one_second
        (1.0001,),  # 1.0001_second
    ],
    ids=[
        "zero_second",
        "one_second",
        "1.0001_second",
    ],
)
def test_split_and_join_with_silence(duration):
    duration = 1.0
    sampling_rate = 10
    sample_width = 2
    channels = 1

    regions = split(
        input="tests/data/test_split_10HZ_mono.raw",
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        analysis_window=0.1,
        sr=sampling_rate,
        sw=sample_width,
        ch=channels,
        eth=50,
    )

    size = round(duration * sampling_rate) * sample_width * channels
    join_data = b"\0" * size
    expected_data = join_data.join(region.data for region in regions)
    expected_region = AudioRegion(
        expected_data, sampling_rate, sample_width, channels
    )

    region_with_silence = split_and_join_with_silence(
        input="tests/data/test_split_10HZ_mono.raw",
        silence_duration=duration,
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        analysis_window=0.1,
        sr=sampling_rate,
        sw=sample_width,
        ch=channels,
        eth=50,
    )
    assert region_with_silence == expected_region


@pytest.mark.parametrize(
    "duration, analysis_window, round_fn, expected, kwargs",
    [
        (0, 1, None, 0, None),  # zero_duration
        (0.3, 0.1, round, 3, None),  # multiple
        (0.35, 0.1, math.ceil, 4, None),  # not_multiple_ceil
        (0.35, 0.1, math.floor, 3, None),  # not_multiple_floor
        (0.05, 0.1, round, 0, None),  # small_duration
        (0.05, 0.1, math.ceil, 1, None),  # small_duration_ceil
        (0.3, 0.1, math.floor, 3, {"epsilon": 1e-6}),  # with_round_error
        (-0.5, 0.1, math.ceil, ValueError, None),  # negative_duration
        (0.5, -0.1, math.ceil, ValueError, None),  # negative_analysis_window
    ],
    ids=[
        "zero_duration",
        "multiple",
        "not_multiple_ceil",
        "not_multiple_floor",
        "small_duration",
        "small_duration_ceil",
        "with_round_error",
        "negative_duration",
        "negative_analysis_window",
    ],
)
def test_duration_to_nb_windows(
    duration, analysis_window, round_fn, expected, kwargs
):
    if expected == ValueError:
        with pytest.raises(ValueError):
            _duration_to_nb_windows(duration, analysis_window, round_fn)
    else:
        if kwargs is None:
            kwargs = {}
        result = _duration_to_nb_windows(
            duration, analysis_window, round_fn, **kwargs
        )
        assert result == expected


@pytest.mark.parametrize(
    "channels, skip, max_read",
    [
        (1, 0, None),  # mono_skip_0_max_read_None
        (1, 3, None),  # mono_skip_3_max_read_None
        (1, 2, -1),  # mono_skip_2_max_read_negative
        (1, 2, 3),  # mono_skip_2_max_read_3
        (2, 0, None),  # stereo_skip_0_max_read_None
        (2, 3, None),  # stereo_skip_3_max_read_None
        (2, 2, -1),  # stereo_skip_2_max_read_negative
        (2, 2, 3),  # stereo_skip_2_max_read_3
    ],
    ids=[
        "mono_skip_0_max_read_None",
        "mono_skip_3_max_read_None",
        "mono_skip_2_max_read_negative",
        "mono_skip_2_max_read_3",
        "stereo_skip_0_max_read_None",
        "stereo_skip_3_max_read_None",
        "stereo_skip_2_max_read_negative",
        "stereo_skip_2_max_read_3",
    ],
)
def test_read_offline(channels, skip, max_read):
    sampling_rate = 10
    sample_width = 2
    mono_or_stereo = "mono" if channels == 1 else "stereo"
    filename = "tests/data/test_split_10HZ_{}.raw".format(mono_or_stereo)
    with open(filename, "rb") as fp:
        data = fp.read()
    onset = round(skip * sampling_rate * sample_width * channels)
    if max_read in (-1, None):
        offset = len(data) + 1
    else:
        offset = onset + round(
            max_read * sampling_rate * sample_width * channels
        )
    expected_data = data[onset:offset]
    read_data, *audio_params = _read_offline(
        filename,
        skip=skip,
        max_read=max_read,
        sr=sampling_rate,
        sw=sample_width,
        ch=channels,
    )
    assert read_data == expected_data
    assert tuple(audio_params) == (sampling_rate, sample_width, channels)


@pytest.mark.parametrize(
    (
        "min_dur, max_dur, max_silence, max_trailing_silence, "
        + "strict_min_dur, kwargs, expected"
    ),
    [
        (
            0.2,
            5,
            0.2,
            None,
            False,
            {"eth": 50},
            [(2, 16), (17, 31), (34, 76)],
        ),  # simple
        (
            0.3,
            2,
            0.2,
            None,
            False,
            {"eth": 50},
            [(2, 16), (17, 31), (34, 54), (54, 74), (74, 76)],
        ),  # short_max_dur
        (3, 5, 0.2, None, False, {"eth": 50}, [(34, 76)]),  # long_min_dur
        (0.2, 80, 10, None, False, {"eth": 50}, [(2, 76)]),  # long_max_silence
        (
            0.2,
            5,
            0.0,
            None,
            False,
            {"eth": 50},
            [(2, 14), (17, 24), (26, 29), (34, 76)],
        ),  # zero_max_silence
        (
            0.2,
            5,
            0.2,
            None,
            False,
            {"energy_threshold": 40},
            [(0, 50), (50, 76)],
        ),  # low_energy_threshold
        (
            0.2,
            5,
            0.2,
            None,
            False,
            {"energy_threshold": 60},
            [],
        ),  # high_energy_threshold
        (
            0.2,
            10,
            0.5,
            0,
            False,
            {"eth": 50},
            [(2, 76)],
        ),  # trim_leading_and_trailing_silence
        (
            0.2,
            5,
            0.2,
            0,
            False,
            {"eth": 50},
            [(2, 14), (17, 29), (34, 76)],
        ),  # drop_trailing_silence
        (
            1.5,
            5,
            0.2,
            0,
            False,
            {"eth": 50},
            [(34, 76)],
        ),  # drop_trailing_silence_2
        (
            0.3,
            2,
            0.2,
            None,
            True,
            {"eth": 50},
            [(2, 16), (17, 31), (34, 54), (54, 74)],
        ),  # strict_min_dur
    ],
    ids=[
        "simple",
        "short_max_dur",
        "long_min_dur",
        "long_max_silence",
        "zero_max_silence",
        "low_energy_threshold",
        "high_energy_threshold",
        "trim_leading_and_trailing_silence",
        "drop_trailing_silence",
        "drop_trailing_silence_2",
        "strict_min_dur",
    ],
)
def test_split_params(
    min_dur,
    max_dur,
    max_silence,
    max_trailing_silence,
    strict_min_dur,
    kwargs,
    expected,
):
    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()

    split_kwargs = dict(
        min_dur=min_dur,
        max_dur=max_dur,
        max_silence=max_silence,
        strict_min_dur=strict_min_dur,
        analysis_window=0.1,
        **kwargs,
    )
    if max_trailing_silence is not None:
        split_kwargs["max_trailing_silence"] = max_trailing_silence

    regions = split(
        data,
        sr=10,
        sw=2,
        ch=1,
        **split_kwargs,
    )

    region = AudioRegion(data, 10, 2, 1)
    regions_ar = region.split(
        **split_kwargs,
    )

    regions = list(regions)
    regions_ar = list(regions_ar)
    err_msg = "Wrong number of regions after split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions))
    assert len(regions) == len(expected), err_msg
    err_msg = "Wrong number of regions after AudioRegion.split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions_ar))
    assert len(regions_ar) == len(expected), err_msg

    sample_width = 2
    for reg, reg_ar, exp in zip(
        regions,
        regions_ar,
        expected,
    ):
        onset, offset = exp
        exp_data = data[onset * sample_width : offset * sample_width]
        assert bytes(reg) == exp_data
        assert reg == reg_ar


@pytest.mark.parametrize(
    "channels, kwargs, expected",
    [
        (2, {}, [(2, 32), (34, 76)]),  # stereo_all_default
        (1, {"max_read": 5}, [(2, 16), (17, 31), (34, 50)]),  # mono_max_read
        (
            1,
            {"mr": 5},
            [(2, 16), (17, 31), (34, 50)],
        ),  # mono_max_read_short_name
        (
            1,
            {"eth": 50, "use_channel": 0},
            [(2, 16), (17, 31), (34, 76)],
        ),  # mono_use_channel_1
        (1, {"eth": 50, "uc": 1}, [(2, 16), (17, 31), (34, 76)]),  # mono_uc_1
        (
            1,
            {"eth": 50, "use_channel": None},
            [(2, 16), (17, 31), (34, 76)],
        ),  # mono_use_channel_None
        (
            2,
            {"eth": 50, "use_channel": 0},
            [(2, 16), (17, 31), (34, 76)],
        ),  # stereo_use_channel_1
        (
            2,
            {"eth": 50},
            [(2, 32), (34, 76)],
        ),  # stereo_use_channel_no_use_channel_given
        (
            2,
            {"eth": 50, "use_channel": -2},
            [(2, 16), (17, 31), (34, 76)],
        ),  # stereo_use_channel_minus_2
        (2, {"eth": 50, "uc": 1}, [(10, 32), (36, 76)]),  # stereo_uc_2
        (2, {"eth": 50, "uc": -1}, [(10, 32), (36, 76)]),  # stereo_uc_minus_1
        (
            1,
            {"eth": 50, "uc": "mix"},
            [(2, 16), (17, 31), (34, 76)],
        ),  # mono_uc_mix
        (
            2,
            {"energy_threshold": 53.5, "use_channel": "mix"},
            [(54, 76)],
        ),  # stereo_use_channel_mix
        (2, {"eth": 52, "uc": "mix"}, [(17, 26), (54, 76)]),  # stereo_uc_mix
        (
            2,
            {"uc": "mix"},
            [(10, 16), (17, 31), (36, 76)],
        ),  # stereo_uc_mix_default_eth
    ],
    ids=[
        "stereo_all_default",
        "mono_max_read",
        "mono_max_read_short_name",
        "mono_use_channel_1",
        "mono_uc_1",
        "mono_use_channel_None",
        "stereo_use_channel_1",
        "stereo_use_channel_no_use_channel_given",
        "stereo_use_channel_minus_2",
        "stereo_uc_2",
        "stereo_uc_minus_1",
        "mono_uc_mix",
        "stereo_use_channel_mix",
        "stereo_uc_mix",
        "stereo_uc_mix_default_eth",
    ],
)
def test_split_kwargs(channels, kwargs, expected):

    mono_or_stereo = "mono" if channels == 1 else "stereo"
    filename = "tests/data/test_split_10HZ_{}.raw".format(mono_or_stereo)
    with open(filename, "rb") as fp:
        data = fp.read()

    regions = split(
        data,
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        analysis_window=0.1,
        sr=10,
        sw=2,
        ch=channels,
        **kwargs,
    )

    region = AudioRegion(data, 10, 2, channels)
    max_read = kwargs.get("max_read", kwargs.get("mr"))
    if max_read is not None:
        region = region.sec[:max_read]
        kwargs.pop("max_read", None)
        kwargs.pop("mr", None)

    regions_ar = region.split(
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        analysis_window=0.1,
        **kwargs,
    )

    regions = list(regions)
    regions_ar = list(regions_ar)
    err_msg = "Wrong number of regions after split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions))
    assert len(regions) == len(expected), err_msg
    err_msg = "Wrong number of regions after AudioRegion.split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions_ar))
    assert len(regions_ar) == len(expected), err_msg

    sample_width = 2
    sample_size_bytes = sample_width * channels
    for reg, reg_ar, exp in zip(
        regions,
        regions_ar,
        expected,
    ):
        onset, offset = exp
        exp_data = data[onset * sample_size_bytes : offset * sample_size_bytes]
        assert len(bytes(reg)) == len(exp_data)
        assert reg == reg_ar


@pytest.mark.parametrize(
    "min_dur, max_dur, max_silence, channels, kwargs, expected",
    [
        (
            0.2,
            5,
            0.2,
            1,
            {"aw": 0.2},
            [(2, 30), (34, 76)],
        ),  # mono_aw_0_2_max_silence_0_2
        (
            0.2,
            5,
            0.3,
            1,
            {"aw": 0.2},
            [(2, 30), (34, 76)],
        ),  # mono_aw_0_2_max_silence_0_3
        (
            0.2,
            5,
            0.4,
            1,
            {"aw": 0.2},
            [(2, 32), (34, 76)],
        ),  # mono_aw_0_2_max_silence_0_4
        (
            0.2,
            5,
            0,
            1,
            {"aw": 0.2},
            [(2, 14), (16, 24), (26, 28), (34, 76)],
        ),  # mono_aw_0_2_max_silence_0
        (0.2, 5, 0.2, 1, {"aw": 0.2}, [(2, 30), (34, 76)]),  # mono_aw_0_2
        (
            0.3,
            5,
            0,
            1,
            {"aw": 0.3},
            [(3, 12), (15, 24), (36, 76)],
        ),  # mono_aw_0_3_max_silence_0
        (
            0.3,
            5,
            0.3,
            1,
            {"aw": 0.3},
            [(3, 27), (36, 76)],
        ),  # mono_aw_0_3_max_silence_0_3
        (
            0.3,
            5,
            0.5,
            1,
            {"aw": 0.3},
            [(3, 27), (36, 76)],
        ),  # mono_aw_0_3_max_silence_0_5
        (
            0.3,
            5,
            0.6,
            1,
            {"aw": 0.3},
            [(3, 30), (36, 76)],
        ),  # mono_aw_0_3_max_silence_0_6
        (
            0.2,
            5,
            0,
            1,
            {"aw": 0.4},
            [(4, 12), (16, 24), (36, 76)],
        ),  # mono_aw_0_4_max_silence_0
        (
            0.2,
            5,
            0.3,
            1,
            {"aw": 0.4},
            [(4, 12), (16, 24), (36, 76)],
        ),  # mono_aw_0_4_max_silence_0_3
        (
            0.2,
            5,
            0.4,
            1,
            {"aw": 0.4},
            [(4, 28), (36, 76)],
        ),  # mono_aw_0_4_max_silence_0_4
        (
            0.2,
            5,
            0.2,
            2,
            {"analysis_window": 0.2},
            [(2, 32), (34, 76)],
        ),  # stereo_uc_None_analysis_window_0_2
        (
            0.2,
            5,
            0.2,
            2,
            {"uc": None, "analysis_window": 0.2},
            [(2, 32), (34, 76)],
        ),  # stereo_uc_any_analysis_window_0_2
        (
            0.2,
            5,
            0.2,
            2,
            {"use_channel": None, "analysis_window": 0.3},
            [(3, 30), (36, 76)],
        ),  # stereo_use_channel_None_aw_0_3_max_silence_0_2
        (
            0.2,
            5,
            0.3,
            2,
            {"use_channel": "any", "analysis_window": 0.3},
            [(3, 33), (36, 76)],
        ),  # stereo_use_channel_any_aw_0_3_max_silence_0_3
        (
            0.2,
            5,
            0.2,
            2,
            {"use_channel": None, "analysis_window": 0.4},
            [(4, 28), (36, 76)],
        ),  # stereo_use_channel_None_aw_0_4_max_silence_0_2
        (
            0.2,
            5,
            0.4,
            2,
            {"use_channel": "any", "analysis_window": 0.4},
            [(4, 32), (36, 76)],
        ),  # stereo_use_channel_any_aw_0_3_max_silence_0_4
        (
            0.2,
            5,
            0.2,
            2,
            {"uc": 0, "analysis_window": 0.2},
            [(2, 30), (34, 76)],
        ),  # stereo_uc_0_analysis_window_0_2
        (
            0.2,
            5,
            0.2,
            2,
            {"uc": 1, "analysis_window": 0.2},
            [(10, 32), (36, 76)],
        ),  # stereo_uc_1_analysis_window_0_2
        (
            0.2,
            5,
            0,
            2,
            {"uc": "mix", "analysis_window": 0.1},
            [(10, 14), (17, 24), (26, 29), (36, 76)],
        ),  # stereo_uc_mix_aw_0_1_max_silence_0
        (
            0.2,
            5,
            0.1,
            2,
            {"uc": "mix", "analysis_window": 0.1},
            [(10, 15), (17, 25), (26, 30), (36, 76)],
        ),  # stereo_uc_mix_aw_0_1_max_silence_0_1
        (
            0.2,
            5,
            0.2,
            2,
            {"uc": "mix", "analysis_window": 0.1},
            [(10, 16), (17, 31), (36, 76)],
        ),  # stereo_uc_mix_aw_0_1_max_silence_0_2
        (
            0.2,
            5,
            0.3,
            2,
            {"uc": "mix", "analysis_window": 0.1},
            [(10, 32), (36, 76)],
        ),  # stereo_uc_mix_aw_0_1_max_silence_0_3
        (
            0.3,
            5,
            0,
            2,
            {"uc": "avg", "analysis_window": 0.2},
            [(10, 14), (16, 24), (36, 76)],
        ),  # stereo_uc_avg_aw_0_2_max_silence_0_min_dur_0_3
        (
            0.41,
            5,
            0,
            2,
            {"uc": "average", "analysis_window": 0.2},
            [(16, 24), (36, 76)],
        ),  # stereo_uc_average_aw_0_2_max_silence_0_min_dur_0_41
        (
            0.2,
            5,
            0.1,
            2,
            {"uc": "mix", "analysis_window": 0.2},
            [(10, 14), (16, 24), (26, 28), (36, 76)],
        ),  # stereo_uc_mix_aw_0_2_max_silence_0_1
        (
            0.2,
            5,
            0.2,
            2,
            {"uc": "mix", "analysis_window": 0.2},
            [(10, 30), (36, 76)],
        ),  # stereo_uc_mix_aw_0_2_max_silence_0_2
        (
            0.2,
            5,
            0.4,
            2,
            {"uc": "mix", "analysis_window": 0.2},
            [(10, 32), (36, 76)],
        ),  # stereo_uc_mix_aw_0_2_max_silence_0_4
        (
            0.2,
            5,
            0.5,
            2,
            {"uc": "mix", "analysis_window": 0.2},
            [(10, 32), (36, 76)],
        ),  # stereo_uc_mix_aw_0_2_max_silence_0_5
        (
            0.2,
            5,
            0.6,
            2,
            {"uc": "mix", "analysis_window": 0.2},
            [(10, 34), (36, 76)],
        ),  # stereo_uc_mix_aw_0_2_max_silence_0_6
        (
            0.2,
            5,
            0,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 24), (27, 30), (36, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0
        (
            0.4,
            5,
            0,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 24), (36, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0_min_dur_0_3
        (
            0.2,
            5,
            0.6,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 57), (57, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0_6
        (
            0.2,
            5.1,
            0.6,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 60), (60, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_1
        (
            0.2,
            5.2,
            0.6,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 60), (60, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_2
        (
            0.2,
            5.3,
            0.6,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 60), (60, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_3
        (
            0.2,
            5.4,
            0.6,
            2,
            {"uc": "mix", "analysis_window": 0.3},
            [(9, 63), (63, 76)],
        ),  # stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_4
        (
            0.2,
            5,
            0,
            2,
            {"uc": "mix", "analysis_window": 0.4},
            [(16, 24), (36, 76)],
        ),  # stereo_uc_mix_aw_0_4_max_silence_0
        (
            0.2,
            5,
            0.3,
            2,
            {"uc": "mix", "analysis_window": 0.4},
            [(16, 24), (36, 76)],
        ),  # stereo_uc_mix_aw_0_4_max_silence_0_3
        (
            0.2,
            5,
            0.4,
            2,
            {"uc": "mix", "analysis_window": 0.4},
            [(16, 28), (36, 76)],
        ),  # stereo_uc_mix_aw_0_4_max_silence_0_4
    ],
    ids=[
        "mono_aw_0_2_max_silence_0_2",
        "mono_aw_0_2_max_silence_0_3",
        "mono_aw_0_2_max_silence_0_4",
        "mono_aw_0_2_max_silence_0",
        "mono_aw_0_2",
        "mono_aw_0_3_max_silence_0",
        "mono_aw_0_3_max_silence_0_3",
        "mono_aw_0_3_max_silence_0_5",
        "mono_aw_0_3_max_silence_0_6",
        "mono_aw_0_4_max_silence_0",
        "mono_aw_0_4_max_silence_0_3",
        "mono_aw_0_4_max_silence_0_4",
        "stereo_uc_None_analysis_window_0_2",
        "stereo_uc_any_analysis_window_0_2",
        "stereo_use_channel_None_aw_0_3_max_silence_0_2",
        "stereo_use_channel_any_aw_0_3_max_silence_0_3",
        "stereo_use_channel_None_aw_0_4_max_silence_0_2",
        "stereo_use_channel_any_aw_0_3_max_silence_0_4",
        "stereo_uc_0_analysis_window_0_2",
        "stereo_uc_1_analysis_window_0_2",
        "stereo_uc_mix_aw_0_1_max_silence_0",
        "stereo_uc_mix_aw_0_1_max_silence_0_1",
        "stereo_uc_mix_aw_0_1_max_silence_0_2",
        "stereo_uc_mix_aw_0_1_max_silence_0_3",
        "stereo_uc_avg_aw_0_2_max_silence_0_min_dur_0_3",
        "stereo_uc_average_aw_0_2_max_silence_0_min_dur_0_41",
        "stereo_uc_mix_aw_0_2_max_silence_0_1",
        "stereo_uc_mix_aw_0_2_max_silence_0_2",
        "stereo_uc_mix_aw_0_2_max_silence_0_4",
        "stereo_uc_mix_aw_0_2_max_silence_0_5",
        "stereo_uc_mix_aw_0_2_max_silence_0_6",
        "stereo_uc_mix_aw_0_3_max_silence_0",
        "stereo_uc_mix_aw_0_3_max_silence_0_min_dur_0_3",
        "stereo_uc_mix_aw_0_3_max_silence_0_6",
        "stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_1",
        "stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_2",
        "stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_3",
        "stereo_uc_mix_aw_0_3_max_silence_0_6_max_dur_5_4",
        "stereo_uc_mix_aw_0_4_max_silence_0",
        "stereo_uc_mix_aw_0_4_max_silence_0_3",
        "stereo_uc_mix_aw_0_4_max_silence_0_4",
    ],
)
def test_split_analysis_window(
    min_dur, max_dur, max_silence, channels, kwargs, expected
):

    mono_or_stereo = "mono" if channels == 1 else "stereo"
    filename = "tests/data/test_split_10HZ_{}.raw".format(mono_or_stereo)
    with open(filename, "rb") as fp:
        data = fp.read()

    regions = split(
        data,
        min_dur=min_dur,
        max_dur=max_dur,
        max_silence=max_silence,
        strict_min_dur=False,
        sr=10,
        sw=2,
        ch=channels,
        eth=49.99,
        **kwargs,
    )

    region = AudioRegion(data, 10, 2, channels)
    regions_ar = region.split(
        min_dur=min_dur,
        max_dur=max_dur,
        max_silence=max_silence,
        strict_min_dur=False,
        eth=49.99,
        **kwargs,
    )

    regions = list(regions)
    regions_ar = list(regions_ar)
    err_msg = "Wrong number of regions after split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions))
    assert len(regions) == len(expected), err_msg
    err_msg = "Wrong number of regions after AudioRegion.split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions_ar))
    assert len(regions_ar) == len(expected), err_msg

    sample_width = 2
    sample_size_bytes = sample_width * channels
    for reg, reg_ar, exp in zip(
        regions,
        regions_ar,
        expected,
    ):
        onset, offset = exp
        exp_data = data[onset * sample_size_bytes : offset * sample_size_bytes]
        assert bytes(reg) == exp_data
        assert reg == reg_ar


def test_split_custom_validator():
    filename = "tests/data/test_split_10HZ_mono.raw"
    with open(filename, "rb") as fp:
        data = fp.read()

    regions = split(
        data,
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        sr=10,
        sw=2,
        ch=1,
        analysis_window=0.1,
        validator=lambda x: to_array(x, sample_width=2, channels=1)[0] >= 320,
    )

    region = AudioRegion(data, 10, 2, 1)
    regions_ar = region.split(
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        analysis_window=0.1,
        validator=lambda x: to_array(x, sample_width=2, channels=1)[0] >= 320,
    )

    expected = [(2, 16), (17, 31), (34, 76)]
    regions = list(regions)
    regions_ar = list(regions_ar)
    err_msg = "Wrong number of regions after split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions))
    assert len(regions) == len(expected), err_msg
    err_msg = "Wrong number of regions after AudioRegion.split, expected: "
    err_msg += "{}, found: {}".format(len(expected), len(regions_ar))
    assert len(regions_ar) == len(expected), err_msg

    sample_size_bytes = 2
    for reg, reg_ar, exp in zip(
        regions,
        regions_ar,
        expected,
    ):
        onset, offset = exp
        exp_data = data[onset * sample_size_bytes : offset * sample_size_bytes]
        assert bytes(reg) == exp_data
        assert reg == reg_ar


@pytest.mark.parametrize(
    "input, kwargs",
    [
        (
            "tests/data/test_split_10HZ_stereo.raw",
            {"audio_format": "raw", "sr": 10, "sw": 2, "ch": 2},
        ),  # filename_audio_format
        (
            "tests/data/test_split_10HZ_stereo.raw",
            {"fmt": "raw", "sr": 10, "sw": 2, "ch": 2},
        ),  # filename_audio_format_short_name
        (
            "tests/data/test_split_10HZ_stereo.raw",
            {"sr": 10, "sw": 2, "ch": 2},
        ),  # filename_no_audio_format
        (
            "tests/data/test_split_10HZ_stereo.raw",
            {"sampling_rate": 10, "sample_width": 2, "channels": 2},
        ),  # filename_no_long_audio_params
        (
            open("tests/data/test_split_10HZ_stereo.raw", "rb").read(),
            {"sr": 10, "sw": 2, "ch": 2},
        ),  # bytes_
        (
            AudioReader(
                "tests/data/test_split_10HZ_stereo.raw",
                sr=10,
                sw=2,
                ch=2,
                block_dur=0.1,
            ),
            {},
        ),  # audio_reader
        (
            AudioRegion(
                open("tests/data/test_split_10HZ_stereo.raw", "rb").read(),
                10,
                2,
                2,
            ),
            {},
        ),  # audio_region
        (
            get_audio_source(
                "tests/data/test_split_10HZ_stereo.raw", sr=10, sw=2, ch=2
            ),
            {},
        ),  # audio_source
    ],
    ids=[
        "filename_audio_format",
        "filename_audio_format_short_name",
        "filename_no_audio_format",
        "filename_no_long_audio_params",
        "bytes_",
        "audio_reader",
        "audio_region",
        "audio_source",
    ],
)
def test_split_input_type(input, kwargs):

    with open("tests/data/test_split_10HZ_stereo.raw", "rb") as fp:
        data = fp.read()

    regions = split(
        input,
        min_dur=0.2,
        max_dur=5,
        max_silence=0.2,
        strict_min_dur=False,
        analysis_window=0.1,
        **kwargs,
    )
    regions = list(regions)
    expected = [(2, 32), (34, 76)]
    sample_width = 2
    err_msg = "Wrong number of regions after split, expected: "
    err_msg += "{}, found: {}".format(expected, regions)
    assert len(regions) == len(expected), err_msg
    for reg, exp in zip(
        regions,
        expected,
    ):
        onset, offset = exp
        exp_data = data[onset * sample_width * 2 : offset * sample_width * 2]
        assert bytes(reg) == exp_data


@pytest.mark.parametrize(
    "min_dur, max_dur, analysis_window",
    [
        (0.5, 0.4, 0.1),
        (0.44, 0.49, 0.1),
    ],
    ids=[
        "min_dur_greater_than_max_dur",
        "durations_OK_but_wrong_number_of_analysis_windows",
    ],
)
def test_split_wrong_min_max_dur(min_dur, max_dur, analysis_window):

    with pytest.raises(ValueError) as val_err:
        split(
            b"0" * 16,
            min_dur=min_dur,
            max_dur=max_dur,
            max_silence=0.2,
            sr=16000,
            sw=1,
            ch=1,
            analysis_window=analysis_window,
        )

    err_msg = "'min_dur' ({0} sec.) results in {1} analysis "
    err_msg += "window(s) ({1} == ceil({0} / {2})) which is "
    err_msg += "higher than the number of analysis window(s) for "
    err_msg += "'max_dur' ({3} == floor({4} / {2}))"

    err_msg = err_msg.format(
        min_dur,
        math.ceil(min_dur / analysis_window),
        analysis_window,
        math.floor(max_dur / analysis_window),
        max_dur,
    )
    assert err_msg == str(val_err.value)


@pytest.mark.parametrize(
    "max_silence, max_dur, analysis_window",
    [
        (0.5, 0.5, 0.1),  # max_silence_equals_max_dur
        (0.5, 0.4, 0.1),  # max_silence_greater_than_max_dur
        (0.44, 0.49, 0.1),  # durations_OK_but_wrong_number_of_analysis_windows
    ],
    ids=[
        "max_silence_equals_max_dur",
        "max_silence_greater_than_max_dur",
        "durations_OK_but_wrong_number_of_analysis_windows",
    ],
)
def test_split_wrong_max_silence_max_dur(max_silence, max_dur, analysis_window):

    with pytest.raises(ValueError) as val_err:
        split(
            b"0" * 16,
            min_dur=0.2,
            max_dur=max_dur,
            max_silence=max_silence,
            sr=16000,
            sw=1,
            ch=1,
            analysis_window=analysis_window,
        )

    err_msg = "'max_silence' ({0} sec.) results in {1} analysis "
    err_msg += "window(s) ({1} == floor({0} / {2})) which is "
    err_msg += "higher or equal to the number of analysis window(s) for "
    err_msg += "'max_dur' ({3} == floor({4} / {2}))"

    err_msg = err_msg.format(
        max_silence,
        math.floor(max_silence / analysis_window),
        analysis_window,
        math.floor(max_dur / analysis_window),
        max_dur,
    )
    assert err_msg == str(val_err.value)


@pytest.mark.parametrize(
    "wrong_param",
    [
        {"min_dur": -1},  # negative_min_dur
        {"min_dur": 0},  # zero_min_dur
        {"max_dur": -1},  # negative_max_dur
        {"max_dur": 0},  # zero_max_dur
        {"max_silence": -1},  # negative_max_silence
        {"analysis_window": 0},  # zero_analysis_window
        {"analysis_window": -1},  # negative_analysis_window
    ],
    ids=[
        "negative_min_dur",
        "zero_min_dur",
        "negative_max_dur",
        "zero_max_dur",
        "negative_max_silence",
        "zero_analysis_window",
        "negative_analysis_window",
    ],
)
def test_split_negative_temporal_params(wrong_param):

    params = {
        "min_dur": 0.2,
        "max_dur": 0.5,
        "max_silence": 0.1,
        "analysis_window": 0.1,
    }
    params.update(wrong_param)
    with pytest.raises(ValueError) as val_err:
        split(None, **params)

    name = set(wrong_param).pop()
    value = wrong_param[name]
    err_msg = "'{}' ({}) must be >{} 0".format(
        name, value, "=" if name == "max_silence" else ""
    )
    assert err_msg == str(val_err.value)


@pytest.mark.parametrize(
    "max_dur",
    [None, float("inf")],
    ids=["None", "inf"],
)
def test_split_max_dur_unlimited(max_dur):
    """With max_dur=None or inf, no truncation occurs."""
    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()

    # With max_dur=2 and max_silence=0.2, the "short_max_dur" case gives
    # 5 regions due to truncation:
    #   [(2,16), (17,31), (34,54), (54,74), (74,76)]
    # The third region (34,76) with duration 4.2s is truncated at max_dur=2.
    # With unlimited max_dur, no truncation happens so we get 3 regions:
    #   [(2,16), (17,31), (34,76)]
    regions = list(
        split(
            data,
            min_dur=0.3,
            max_dur=max_dur,
            max_silence=0.2,
            analysis_window=0.1,
            sr=10,
            sw=2,
            ch=1,
            eth=50,
        )
    )
    expected = [(2, 16), (17, 31), (34, 76)]
    assert len(regions) == len(expected)
    sample_width = 2
    for reg, (onset, offset) in zip(
        regions,
        expected,
    ):
        exp_data = data[onset * sample_width : offset * sample_width]
        assert bytes(reg) == exp_data


def test_split_max_dur_none_via_audio_region():
    """AudioRegion.split also accepts max_dur=None."""
    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()

    region = AudioRegion(data, 10, 2, 1)
    regions = list(
        region.split(
            min_dur=0.2,
            max_dur=None,
            max_silence=0.2,
            analysis_window=0.1,
            eth=50,
        )
    )
    regions_from_split = list(
        split(
            data,
            min_dur=0.2,
            max_dur=None,
            max_silence=0.2,
            analysis_window=0.1,
            sr=10,
            sw=2,
            ch=1,
            eth=50,
        )
    )
    assert len(regions) == len(regions_from_split)
    for r1, r2 in zip(
        regions,
        regions_from_split,
    ):
        assert r1 == r2


def test_split_too_small_analysis_window():
    with pytest.raises(ValueError) as val_err:
        split(b"", sr=10, sw=1, ch=1, analysis_window=0.09)
    err_msg = "Too small 'analysis_window' (0.09) for sampling rate (10)."
    err_msg += " Analysis window should at least be 1/10 to cover one "
    err_msg += "data sample"
    assert err_msg == str(val_err.value)


def test_split_and_plot():

    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()

    region = AudioRegion(data, 10, 2, 1)
    with patch("auditok.audio.plot") as patch_fn:
        regions = region.split_and_plot(
            min_dur=0.2,
            max_dur=5,
            max_silence=0.2,
            strict_min_dur=False,
            analysis_window=0.1,
            sr=10,
            sw=2,
            ch=1,
            eth=50,
        )
    assert patch_fn.called
    expected = [(2, 16), (17, 31), (34, 76)]
    sample_width = 2
    expected_regions = []
    for onset, offset in expected:
        onset *= sample_width
        offset *= sample_width
        expected_regions.append(AudioRegion(data[onset:offset], 10, 2, 1))
    assert regions == expected_regions


def test_split_and_plot_interactive_in_notebook():
    """interactive=True in a notebook calls display_interactive, not plot."""
    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()
    region = AudioRegion(data, 10, 2, 1)
    with patch("auditok.widget._in_notebook", return_value=True), patch(
        "auditok.widget.display_interactive"
    ) as mock_display, patch("auditok.audio.plot") as mock_plot:
        regions = region.split_and_plot(
            min_dur=0.2,
            max_dur=5,
            max_silence=0.2,
            analysis_window=0.1,
            interactive=True,
            sr=10,
            sw=2,
            ch=1,
            eth=50,
        )
    assert mock_display.called
    assert not mock_plot.called
    assert len(regions) == 3


def test_split_and_plot_interactive_not_in_notebook():
    """interactive=True outside a notebook falls back to matplotlib plot."""
    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()
    region = AudioRegion(data, 10, 2, 1)
    with patch("auditok.widget._in_notebook", return_value=False), patch(
        "auditok.widget.display_interactive"
    ) as mock_display, patch("auditok.audio.plot") as mock_plot:
        regions = region.split_and_plot(
            min_dur=0.2,
            max_dur=5,
            max_silence=0.2,
            analysis_window=0.1,
            interactive=True,
            sr=10,
            sw=2,
            ch=1,
            eth=50,
        )
    assert not mock_display.called
    assert mock_plot.called
    assert len(regions) == 3


def test_split_exception():
    with open("tests/data/test_split_10HZ_mono.raw", "rb") as fp:
        data = fp.read()
        region = AudioRegion(data, 10, 2, 1)

    with pytest.raises(RuntimeWarning):
        # max_read is not accepted when calling AudioRegion.split
        region.split(max_read=2)


@pytest.mark.parametrize(
    (
        "data, start, sampling_rate, sample_width, channels, expected_end, "
        + "expected_duration_s, expected_duration_ms"
    ),
    [
        (b"\0" * 8000, 0, 8000, 1, 1, 1, 1, 1000),  # simple
        (
            b"\0" * 7992,
            0,
            8000,
            1,
            1,
            0.999,
            0.999,
            999,
        ),  # one_ms_less_than_1_sec
        (
            b"\0" * 7994,
            0,
            8000,
            1,
            1,
            0.99925,
            0.99925,
            999,
        ),  # tree_quarter_ms_less_than_1_sec
        (
            b"\0" * 7996,
            0,
            8000,
            1,
            1,
            0.9995,
            0.9995,
            1000,
        ),  # half_ms_less_than_1_sec
        (
            b"\0" * 7998,
            0,
            8000,
            1,
            1,
            0.99975,
            0.99975,
            1000,
        ),  # quarter_ms_less_than_1_sec
        (b"\0" * 8000 * 2, 0, 8000, 2, 1, 1, 1, 1000),  # simple_sample_width_2
        (b"\0" * 8000 * 2, 0, 8000, 1, 2, 1, 1, 1000),  # simple_stereo
        (b"\0" * 8000 * 5, 0, 8000, 1, 5, 1, 1, 1000),  # simple_multichannel
        (
            b"\0" * 8000 * 2 * 5,
            0,
            8000,
            2,
            5,
            1,
            1,
            1000,
        ),  # simple_sample_width_2_multichannel
        (
            b"\0" * 7992 * 2 * 5,
            0,
            8000,
            2,
            5,
            0.999,
            0.999,
            999,
        ),  # one_ms_less_than_1s_sw_2_multichannel
        (
            b"\0" * 7994 * 2 * 5,
            0,
            8000,
            2,
            5,
            0.99925,
            0.99925,
            999,
        ),  # tree_qrt_ms_lt_1_s_sw_2_multichannel
        (
            b"\0" * 7996 * 2 * 5,
            0,
            8000,
            2,
            5,
            0.9995,
            0.9995,
            1000,
        ),  # half_ms_lt_1s_sw_2_multichannel
        (
            b"\0" * 7998 * 2 * 5,
            0,
            8000,
            2,
            5,
            0.99975,
            0.99975,
            1000,
        ),  # quarter_ms_lt_1s_sw_2_multichannel
        (
            b"\0" * int(8000 * 1.33),
            2.7,
            8000,
            1,
            1,
            4.03,
            1.33,
            1330,
        ),  # arbitrary_length_1
        (
            b"\0" * int(8000 * 0.476),
            11.568,
            8000,
            1,
            1,
            12.044,
            0.476,
            476,
        ),  # arbitrary_length_2
        (
            b"\0" * int(8000 * 1.711) * 2 * 3,
            9.415,
            8000,
            2,
            3,
            11.126,
            1.711,
            1711,
        ),  # arbitrary_length_sw_2_multichannel
        (
            b"\0" * int(3172 * 1.318),
            17.236,
            3172,
            1,
            1,
            17.236 + int(3172 * 1.318) / 3172,
            int(3172 * 1.318) / 3172,
            1318,
        ),  # arbitrary_sampling_rate
        (
            b"\0" * int(11317 * 0.716) * 2 * 3,
            18.811,
            11317,
            2,
            3,
            18.811 + int(11317 * 0.716) / 11317,
            int(11317 * 0.716) / 11317,
            716,
        ),  # arbitrary_sr_sw_2_multichannel
    ],
    ids=[
        "simple",
        "one_ms_less_than_1_sec",
        "tree_quarter_ms_less_than_1_sec",
        "half_ms_less_than_1_sec",
        "quarter_ms_less_than_1_sec",
        "simple_sample_width_2",
        "simple_stereo",
        "simple_multichannel",
        "simple_sample_width_2_multichannel",
        "one_ms_less_than_1s_sw_2_multichannel",
        "tree_qrt_ms_lt_1_s_sw_2_multichannel",
        "half_ms_lt_1s_sw_2_multichannel",
        "quarter_ms_lt_1s_sw_2_multichannel",
        "arbitrary_length_1",
        "arbitrary_length_2",
        "arbitrary_length_sw_2_multichannel",
        "arbitrary_sampling_rate",
        "arbitrary_sr_sw_2_multichannel",
    ],
)
def test_creation(
    data,
    start,
    sampling_rate,
    sample_width,
    channels,
    expected_end,
    expected_duration_s,
    expected_duration_ms,
):
    region = AudioRegion(data, sampling_rate, sample_width, channels, start)
    assert region.sampling_rate == sampling_rate
    assert region.sr == sampling_rate
    assert region.sample_width == sample_width
    assert region.sw == sample_width
    assert region.channels == channels
    assert region.ch == channels
    assert region.start == start
    assert region.end == expected_end
    assert region.duration == expected_duration_s
    assert len(region.ms) == expected_duration_ms
    assert bytes(region) == data


def test_creation_invalid_data_exception():
    with pytest.raises(AudioParameterError) as audio_param_err:
        _ = AudioRegion(
            data=b"ABCDEFGHI", sampling_rate=8, sample_width=2, channels=1
        )
    assert str(audio_param_err.value) == (
        "The length of audio data must be an integer "
        "multiple of `sample_width * channels`"
    )


def test_repr_html():
    """_repr_html_ returns an HTML audio player for Jupyter notebooks."""
    import base64
    import io
    import wave

    data = b"\x00\x01" * 160  # 10ms at 16kHz, 16-bit mono
    region = AudioRegion(data, sampling_rate=16000, sample_width=2, channels=1)
    html = region._repr_html_()

    assert "<audio controls" in html
    assert "AudioRegion" in html
    assert "0.010s" in html
    assert "16000 Hz" in html
    assert "16‑bit" in html
    assert "1 ch" in html
    assert 'type="audio/wav"' in html

    # Extract base64 data and verify it's a valid WAV
    prefix = "data:audio/wav;base64,"
    start = html.index(prefix) + len(prefix)
    end = html.index('"', start)
    wav_bytes = base64.b64decode(html[start:end])
    buf = io.BytesIO(wav_bytes)
    with wave.open(buf, "rb") as wf:
        assert wf.getframerate() == 16000
        assert wf.getsampwidth() == 2
        assert wf.getnchannels() == 1
        assert wf.readframes(160) == data


def test_repr_html_with_start_and_end():
    """_repr_html_ shows start and end times when start is set."""
    data = b"\x00\x01" * 16000  # 1s at 16kHz, 16-bit mono
    region = AudioRegion(
        data, sampling_rate=16000, sample_width=2, channels=1, start=2.5
    )
    html = region._repr_html_()
    assert "2.500s" in html  # start
    assert "3.500s" in html  # end = 2.5 + 1.0


@pytest.mark.parametrize(
    "skip, max_read, channels",
    [
        (0, -1, 1),  # no_skip_read_all
        (0, -1, 2),  # no_skip_read_all_stereo
        (2, -1, 1),  # skip_2_read_all
        (2, None, 1),  # skip_2_read_all_None
        (2, 3, 1),  # skip_2_read_3
        (2, 3.5, 2),  # skip_2_read_3_5_stereo
        (2.4, 3.5, 2),  # skip_2_4_read_3_5_stereo
    ],
    ids=[
        "no_skip_read_all",
        "no_skip_read_all_stereo",
        "skip_2_read_all",
        "skip_2_read_all_None",
        "skip_2_read_3",
        "skip_2_read_3_5_stereo",
        "skip_2_4_read_3_5_stereo",
    ],
)
def test_load_AudioRegion(skip, max_read, channels):
    sampling_rate = 10
    sample_width = 2
    filename = "tests/data/test_split_10HZ_{}.raw"
    filename = filename.format("mono" if channels == 1 else "stereo")
    region = AudioRegion.load(
        filename,
        skip=skip,
        max_read=max_read,
        sr=sampling_rate,
        sw=sample_width,
        ch=channels,
    )
    with open(filename, "rb") as fp:
        fp.read(round(skip * sampling_rate * sample_width * channels))
        if max_read is None or max_read < 0:
            to_read = -1
        else:
            to_read = round(max_read * sampling_rate * sample_width * channels)
        expected = fp.read(to_read)
    assert bytes(region) == expected


def test_load_from_microphone():
    with patch("auditok.io.AudioDeviceSource") as patch_audio_source:
        with patch("auditok.audio.AudioReader.read") as patch_reader:
            patch_reader.return_value = None
            with patch(
                "auditok.audio.AudioRegion.__init__"
            ) as patch_AudioRegion:
                patch_AudioRegion.return_value = None
                AudioRegion.load(None, skip=0, max_read=5, sr=16000, sw=2, ch=1)
    assert patch_audio_source.called
    assert patch_reader.called
    assert patch_AudioRegion.called


@pytest.mark.parametrize(
    "max_read",
    [
        None,  # None
        -1,  # negative
    ],
    ids=[
        "None",
        "negative",
    ],
)
def test_load_from_microphone_without_max_read_exception(max_read):
    with pytest.raises(ValueError) as val_err:
        AudioRegion.load(None, max_read=max_read, sr=16000, sw=2, ch=1)
    assert str(val_err.value) == (
        "'max_read' should not be None when reading from microphone"
    )


def test_load_from_microphone_with_nonzero_skip_exception():
    with pytest.raises(ValueError) as val_err:
        AudioRegion.load(None, skip=1, max_read=5, sr=16000, sw=2, ch=1)
    assert str(val_err.value) == (
        "'skip' should be 0 when reading from microphone"
    )


@pytest.mark.parametrize(
    "format, start, expected",
    [
        ("output.wav", 1.230, "output.wav"),  # simple
        ("output_{start:g}.wav", 1.230, "output_1.23.wav"),  # start
        ("output_{start}.wav", 1.233712, "output_1.233712.wav"),  # start_2
        (
            "output_{start:.2f}.wav",
            1.2300001,
            "output_1.23.wav",
        ),  # start_3
        (
            "output_{start:.3f}.wav",
            1.233712,
            "output_1.234.wav",
        ),  # start_4
        (
            "output_{start:.8f}.wav",
            1.233712,
            "output_1.23371200.wav",
        ),  # start_5
        (
            "output_{start}_{end}_{duration}.wav",
            1.455,
            "output_1.455_2.455_1.0.wav",
        ),  # start_end_duration
        (
            "output_{start}_{end}_{duration}.wav",
            1.455321,
            "output_1.455321_2.455321_1.0.wav",
        ),  # start_end_duration_2
    ],
    ids=[
        "simple",
        "start",
        "start_2",
        "start_3",
        "start_4",
        "start_5",
        "start_end_duration",
        "start_end_duration_2",
    ],
)
def test_save(format, start, expected):
    with TemporaryDirectory() as tmpdir:
        region = AudioRegion(b"0" * 160, 160, 1, 1, start)
        format = os.path.join(tmpdir, format)
        filename = region.save(format)[len(tmpdir) + 1 :]
        assert filename == expected


def test_save_file_exists_exception():
    with TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "output.wav")
        open(filename, "w").close()
        region = AudioRegion(b"0" * 160, 160, 1, 1)
        with pytest.raises(FileExistsError):
            region.save(filename, exists_ok=False)

        with pytest.raises(FileExistsError):
            region.save(Path(filename), exists_ok=False)


def _read_audio_info_ffprobe(filepath):
    import json
    import subprocess

    cmd = [
        "ffprobe",
        "-v",
        "quiet",
        "-print_format",
        "json",
        "-show_streams",
        "-select_streams",
        "a:0",
        str(filepath),
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    info = json.loads(result.stdout)
    stream = info["streams"][0]
    return {
        "sample_rate": int(stream["sample_rate"]),
        "channels": int(stream["channels"]),
        "codec_name": stream["codec_name"],
    }


@requires_ffmpeg
@pytest.mark.parametrize(
    "ext, expected_codec",
    [
        (".ogg", "vorbis"),
        (".mp3", "mp3"),
        (".flac", "flac"),
    ],
    ids=["ogg", "mp3", "flac"],
)
def test_save_compressed_formats(ext, expected_codec):
    region = AudioRegion(
        b"\0" * 32000, sampling_rate=16000, sample_width=2, channels=1
    )
    with TemporaryDirectory() as tmpdir:
        filepath = os.path.join(tmpdir, "audio" + ext)
        region.save(filepath)
        assert os.path.getsize(filepath) > 0
        info = _read_audio_info_ffprobe(filepath)
        assert info["sample_rate"] == 16000
        assert info["channels"] == 1


@requires_ffmpeg
def test_save_with_audio_codec():
    region = AudioRegion(
        b"\0" * 32000, sampling_rate=16000, sample_width=2, channels=1
    )
    with TemporaryDirectory() as tmpdir:
        filepath = os.path.join(tmpdir, "audio.ogg")
        region.save(filepath, audio_codec="libvorbis")
        info = _read_audio_info_ffprobe(filepath)
        assert info["codec_name"] == "vorbis"


@requires_ffmpeg
def test_save_with_audio_bitrate():
    region = AudioRegion(
        b"\0" * 32000, sampling_rate=16000, sample_width=2, channels=1
    )
    with TemporaryDirectory() as tmpdir:
        filepath = os.path.join(tmpdir, "audio.mp3")
        region.save(filepath, audio_bitrate="192k")
        assert os.path.getsize(filepath) > 0


@requires_ffmpeg
def test_save_with_audio_quality():
    region = AudioRegion(
        b"\0" * 32000, sampling_rate=16000, sample_width=2, channels=1
    )
    with TemporaryDirectory() as tmpdir:
        filepath = os.path.join(tmpdir, "audio.mp3")
        region.save(filepath, audio_quality="0")
        assert os.path.getsize(filepath) > 0


@requires_ffmpeg
def test_save_with_ffmpeg_extra_args():
    region = AudioRegion(
        b"\0" * 32000, sampling_rate=16000, sample_width=2, channels=1
    )
    with TemporaryDirectory() as tmpdir:
        filepath = os.path.join(tmpdir, "audio.mp3")
        region.save(filepath, ffmpeg_extra_args=["-ar", "8000"])
        info = _read_audio_info_ffprobe(filepath)
        assert info["sample_rate"] == 8000


@requires_ffmpeg
def test_save_roundtrip_flac():
    """Save as FLAC and read back — FLAC is lossless so data should match."""
    data = b"\0" * 32000
    region = AudioRegion(data, sampling_rate=16000, sample_width=2, channels=1)
    with TemporaryDirectory() as tmpdir:
        filepath = os.path.join(tmpdir, "audio.flac")
        region.save(filepath)
        loaded = AudioRegion.load(filepath)
        assert bytes(loaded) == data


@pytest.mark.parametrize(
    "sampling_rate, sample_width, channels",
    [
        (16000, 1, 1),  # mono_16K_1byte
        (16000, 2, 1),  # mono_16K_2byte
        (44100, 2, 2),  # stereo_44100_2byte
        (44100, 2, 3),  # 3channel_44100_2byte
    ],
    ids=[
        "mono_16K_1byte",
        "mono_16K_2byte",
        "stereo_44100_2byte",
        "3channel_44100_2byte",
    ],
)
def test_join(sampling_rate, sample_width, channels):
    duration = 1
    size = int(duration * sampling_rate * sample_width * channels)
    glue_data = b"\0" * size
    regions_data = [
        b"\1" * int(size * 1.5),
        b"\2" * int(size * 0.5),
        b"\3" * int(size * 0.75),
    ]

    glue_region = AudioRegion(glue_data, sampling_rate, sample_width, channels)
    regions = [
        AudioRegion(data, sampling_rate, sample_width, channels)
        for data in regions_data
    ]
    joined = glue_region.join(regions)
    assert joined.data == glue_data.join(regions_data)
    assert joined.duration == duration * 2 + 1.5 + 0.5 + 0.75


@pytest.mark.parametrize(
    "sampling_rate, sample_width, channels",
    [
        (32000, 1, 1),  # different_sampling_rate
        (16000, 2, 1),  # different_sample_width
        (16000, 1, 2),  # different_channels
    ],
    ids=[
        "different_sampling_rate",
        "different_sample_width",
        "different_channels",
    ],
)
def test_join_exception(sampling_rate, sample_width, channels):

    glue_sampling_rate = 16000
    glue_sample_width = 1
    glue_channels = 1

    duration = 1
    size = int(
        duration * glue_sampling_rate * glue_sample_width * glue_channels
    )
    glue_data = b"\0" * size
    glue_region = AudioRegion(
        glue_data, glue_sampling_rate, glue_sample_width, glue_channels
    )

    size = int(duration * sampling_rate * sample_width * channels)
    regions_data = [
        b"\1" * int(size * 1.5),
        b"\2" * int(size * 0.5),
        b"\3" * int(size * 0.75),
    ]
    regions = [
        AudioRegion(data, sampling_rate, sample_width, channels)
        for data in regions_data
    ]

    with pytest.raises(AudioParameterError):
        glue_region.join(regions)


@pytest.mark.parametrize(
    "region, slice_, expected_data",
    [
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(0, 500),
            b"a" * 80,  # first_half
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(500, None),
            b"b" * 80,  # second_half
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(-500, None),
            b"b" * 80,  # second_half_negative
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(200, 750),
            b"a" * 48 + b"b" * 40,  # middle
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(-800, -250),
            b"a" * 48 + b"b" * 40,  # middle_negative
        ),
        (
            AudioRegion(b"a" * 160 + b"b" * 160, 160, 2, 1),
            slice(200, 750),
            b"a" * 96 + b"b" * 80,  # middle_sw2
        ),
        (
            AudioRegion(b"a" * 160 + b"b" * 160, 160, 1, 2),
            slice(200, 750),
            b"a" * 96 + b"b" * 80,  # middle_ch2
        ),
        (
            AudioRegion(b"a" * 320 + b"b" * 320, 160, 2, 2),
            slice(200, 750),
            b"a" * 192 + b"b" * 160,  # middle_sw2_ch2
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(1, None),
            b"a" * (4000 - 8) + b"b" * 4000,  # but_first_sample
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(-999, None),
            b"a" * (4000 - 8) + b"b" * 4000,  # but_first_sample_negative
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(0, 999),
            b"a" * 4000 + b"b" * (4000 - 8),  # but_last_sample
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(0, -1),
            b"a" * 4000 + b"b" * (4000 - 8),  # but_last_sample_negative
        ),
        (
            AudioRegion(b"a" * 160, 160, 1, 1),
            slice(-5000, None),
            b"a" * 160,  # big_negative_start
        ),
        (
            AudioRegion(b"a" * 160, 160, 1, 1),
            slice(None, -1500),
            b"",  # big_negative_stop
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(0, 0),
            b"",  # empty
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(200, 100),
            b"",  # empty_start_stop_reversed
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(2000, 3000),
            b"",  # empty_big_positive_start
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(-100, -200),
            b"",  # empty_negative_reversed
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(0, -2000),
            b"",  # empty_big_negative_stop
        ),
        (
            AudioRegion(b"a" * 124 + b"b" * 376, 1234, 1, 1),
            slice(100, 200),
            b"a" + b"b" * 123,  # arbitrary_sampling_rate
        ),
    ],
    ids=[
        "first_half",
        "second_half",
        "second_half_negative",
        "middle",
        "middle_negative",
        "middle_sw2",
        "middle_ch2",
        "middle_sw2_ch2",
        "but_first_sample",
        "but_first_sample_negative",
        "but_last_sample",
        "but_last_sample_negative",
        "big_negative_start",
        "big_negative_stop",
        "empty",
        "empty_start_stop_reversed",
        "empty_big_positive_start",
        "empty_negative_reversed",
        "empty_big_negative_stop",
        "arbitrary_sampling_rate",
    ],
)
def test_region_temporal_slicing(region, slice_, expected_data):
    sub_region = region.millis[slice_]
    assert bytes(sub_region) == expected_data
    start_sec = slice_.start / 1000 if slice_.start is not None else None
    stop_sec = slice_.stop / 1000 if slice_.stop is not None else None
    sub_region = region.sec[start_sec:stop_sec]
    assert bytes(sub_region) == expected_data


@pytest.mark.parametrize(
    "region, slice_, time_shift, expected_data",
    [
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(0, 80),
            0,
            b"a" * 80,  # first_half
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(80, None),
            0.5,
            b"b" * 80,  # second_half
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(-80, None),
            0.5,
            b"b" * 80,  # second_half_negative
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(160 // 5, 160 // 4 * 3),
            0.2,
            b"a" * 48 + b"b" * 40,  # middle
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(-160 // 5 * 4, -160 // 4),
            0.2,
            b"a" * 48 + b"b" * 40,  # middle_negative
        ),
        (
            AudioRegion(b"a" * 160 + b"b" * 160, 160, 2, 1),
            slice(160 // 5, 160 // 4 * 3),
            0.2,
            b"a" * 96 + b"b" * 80,  # middle_sw2
        ),
        (
            AudioRegion(b"a" * 160 + b"b" * 160, 160, 1, 2),
            slice(160 // 5, 160 // 4 * 3),
            0.2,
            b"a" * 96 + b"b" * 80,  # middle_ch2
        ),
        (
            AudioRegion(b"a" * 320 + b"b" * 320, 160, 2, 2),
            slice(160 // 5, 160 // 4 * 3),
            0.2,
            b"a" * 192 + b"b" * 160,  # middle_sw2_ch2
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(1, None),
            1 / 8000,
            b"a" * (4000 - 1) + b"b" * 4000,  # but_first_sample
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(-7999, None),
            1 / 8000,
            b"a" * (4000 - 1) + b"b" * 4000,  # but_first_sample_negative
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(0, 7999),
            0,
            b"a" * 4000 + b"b" * (4000 - 1),  # but_last_sample
        ),
        (
            AudioRegion(b"a" * 4000 + b"b" * 4000, 8000, 1, 1),
            slice(0, -1),
            0,
            b"a" * 4000 + b"b" * (4000 - 1),  # but_last_sample_negative
        ),
        (
            AudioRegion(b"a" * 160, 160, 1, 1),
            slice(-1600, None),
            0,
            b"a" * 160,  # big_negative_start
        ),
        (
            AudioRegion(b"a" * 160, 160, 1, 1),
            slice(None, -1600),
            0,
            b"",  # big_negative_stop
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(0, 0),
            0,
            b"",  # empty
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(80, 40),
            0.5,
            b"",  # empty_start_stop_reversed
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(1600, 3000),
            10,
            b"",  # empty_big_positive_start
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(-16, -32),
            0.9,
            b"",  # empty_negative_reversed
        ),
        (
            AudioRegion(b"a" * 80 + b"b" * 80, 160, 1, 1),
            slice(0, -2000),
            0,
            b"",  # empty_big_negative_stop
        ),
        (
            AudioRegion(b"a" * 124 + b"b" * 376, 1235, 1, 1),
            slice(100, 200),
            100 / 1235,
            b"a" * 24 + b"b" * 76,  # arbitrary_sampling_rate
        ),
        (
            AudioRegion(b"a" * 124 + b"b" * 376, 1235, 2, 2),
            slice(25, 50),
            25 / 1235,
            b"a" * 24 + b"b" * 76,  # arbitrary_sampling_rate_middle_sw2_ch2
        ),
    ],
    ids=[
        "first_half",
        "second_half",
        "second_half_negative",
        "middle",
        "middle_negative",
        "middle_sw2",
        "middle_ch2",
        "middle_sw2_ch2",
        "but_first_sample",
        "but_first_sample_negative",
        "but_last_sample",
        "but_last_sample_negative",
        "big_negative_start",
        "big_negative_stop",
        "empty",
        "empty_start_stop_reversed",
        "empty_big_positive_start",
        "empty_negative_reversed",
        "empty_big_negative_stop",
        "arbitrary_sampling_rate",
        "arbitrary_sampling_rate_middle_sw2_ch2",
    ],
)
def test_region_sample_slicing(region, slice_, time_shift, expected_data):
    sub_region = region[slice_]
    assert bytes(sub_region) == expected_data


@pytest.mark.parametrize(
    "sampling_rate, sample_width, channels",
    [
        (8000, 1, 1),  # simple
        (8000, 2, 2),  # stereo_sw_2
        (5413, 2, 3),  # arbitrary_sr_multichannel
    ],
    ids=[
        "simple",
        "stereo_sw_2",
        "arbitrary_sr_multichannel",
    ],
)
def test_concatenation(sampling_rate, sample_width, channels):

    region_1, region_2 = _make_random_length_regions(
        [b"a", b"b"], sampling_rate, sample_width, channels
    )
    expected_duration = region_1.duration + region_2.duration
    expected_data = bytes(region_1) + bytes(region_2)
    concat_region = region_1 + region_2
    assert concat_region.duration == pytest.approx(expected_duration, abs=1e-6)
    assert bytes(concat_region) == expected_data


@pytest.mark.parametrize(
    "sampling_rate, sample_width, channels",
    [
        (8000, 1, 1),  # simple
        (8000, 2, 2),  # stereo_sw_2
        (5413, 2, 3),  # arbitrary_sr_multichannel
    ],
    ids=[
        "simple",
        "stereo_sw_2",
        "arbitrary_sr_multichannel",
    ],
)
def test_concatenation_many(sampling_rate, sample_width, channels):

    regions = _make_random_length_regions(
        [b"a", b"b", b"c"], sampling_rate, sample_width, channels
    )
    expected_duration = sum(r.duration for r in regions)
    expected_data = b"".join(bytes(r) for r in regions)
    concat_region = sum(regions)

    assert concat_region.duration == pytest.approx(expected_duration, abs=1e-6)
    assert bytes(concat_region) == expected_data


def test_concatenation_different_sampling_rate_error():
    region_1 = AudioRegion(b"a" * 100, 8000, 1, 1)
    region_2 = AudioRegion(b"b" * 100, 3000, 1, 1)

    with pytest.raises(AudioParameterError) as val_err:
        region_1 + region_2
    assert str(val_err.value) == (
        "Can only concatenate AudioRegions of the same "
        "sampling rate (8000 != 3000)"  # different_sampling_rate
    )


def test_concatenation_different_sample_width_error():
    region_1 = AudioRegion(b"a" * 100, 8000, 2, 1)
    region_2 = AudioRegion(b"b" * 100, 8000, 4, 1)

    with pytest.raises(AudioParameterError) as val_err:
        region_1 + region_2
    assert str(val_err.value) == (
        "Can only concatenate AudioRegions of the same sample width (2 != 4)"
    )


def test_concatenation_different_number_of_channels_error():
    region_1 = AudioRegion(b"a" * 100, 8000, 1, 1)
    region_2 = AudioRegion(b"b" * 100, 8000, 1, 2)

    with pytest.raises(AudioParameterError) as val_err:
        region_1 + region_2
    assert str(val_err.value) == (
        "Can only concatenate AudioRegions of the same "
        "number of channels (1 != 2)"  # different_number_of_channels
    )


@pytest.mark.parametrize(
    "duration, expected_duration, expected_len, expected_len_ms",
    [
        (0.01, 0.03, 240, 30),  # simple
        (0.00575, 0.01725, 138, 17),  # rounded_len_floor
        (0.00625, 0.01875, 150, 19),  # rounded_len_ceil
    ],
    ids=[
        "simple",
        "rounded_len_floor",
        "rounded_len_ceil",
    ],
)
def test_multiplication(
    duration, expected_duration, expected_len, expected_len_ms
):
    sw = 2
    data = b"0" * int(duration * 8000 * sw)
    region = AudioRegion(data, 8000, sw, 1)
    m_region = 1 * region * 3
    assert bytes(m_region) == data * 3
    assert m_region.sr == 8000
    assert m_region.sw == 2
    assert m_region.ch == 1
    assert m_region.duration == expected_duration
    assert len(m_region) == expected_len
    assert m_region.len == expected_len
    assert m_region.s.len == expected_duration
    assert len(m_region.ms) == expected_len_ms
    assert m_region.ms.len == expected_len_ms


@pytest.mark.parametrize(
    "factor, _type",
    [
        ("x", str),  # string
        (1.4, float),  # float
    ],
    ids=[
        "string",
        "float",
    ],
)
def test_multiplication_non_int(factor, _type):
    with pytest.raises(TypeError) as type_err:
        AudioRegion(b"0" * 80, 8000, 1, 1) * factor
    err_msg = "Can't multiply AudioRegion by a non-int of type '{}'"
    assert err_msg.format(_type) == str(type_err.value)


@pytest.mark.parametrize(
    "data",
    [
        [b"a" * 80, b"b" * 80],  # simple
        [b"a" * 31, b"b" * 31, b"c" * 30],  # extra_samples_1
        [b"a" * 31, b"b" * 30, b"c" * 30],  # extra_samples_2
        [b"a" * 11, b"b" * 11, b"c" * 10, b"c" * 10],  # extra_samples_3
    ],
    ids=[
        "simple",
        "extra_samples_1",
        "extra_samples_2",
        "extra_samples_3",
    ],
)
def test_truediv(data):

    region = AudioRegion(b"".join(data), 80, 1, 1)

    sub_regions = region / len(data)
    for data_i, region in zip(
        data,
        sub_regions,
    ):
        assert len(data_i) == len(bytes(region))


@pytest.mark.parametrize(
    "data, sample_width, channels, expected",
    [
        (b"a" * 10, 1, 1, [97] * 10),  # mono_sw_1
        (b"a" * 10, 2, 1, [24929] * 5),  # mono_sw_2
        (b"a" * 8, 4, 1, [1633771873] * 2),  # mono_sw_4
        (b"ab" * 5, 1, 2, [[97] * 5, [98] * 5]),  # stereo_sw_1
    ],
    ids=[
        "mono_sw_1",
        "mono_sw_2",
        "mono_sw_4",
        "stereo_sw_1",
    ],
)
def test_samples(data, sample_width, channels, expected):

    region = AudioRegion(data, 10, sample_width, channels)
    expected = np.array(expected)
    assert (region.numpy() == expected).all()
    assert (np.array(region) == expected).all()



import os

import pytest

from auditok import DataValidator, StreamTokenizer, StringDataSource


class AValidator(DataValidator):
    def is_valid(self, frame):
        return frame == "A"


@pytest.fixture
def validator():
    return AValidator()


def test_init_min_0_init_max_silence_0(validator):
    tokenizer = StreamTokenizer(
        validator,
        min_length=5,
        max_length=20,
        max_continuous_silence=4,
        init_min=0,
        init_max_silence=0,
        mode=0,
    )

    data_source = StringDataSource("aAaaaAaAaaAaAaaaaaaaAAAAAAAA")
    #                                ^              ^   ^      ^
    #                                2              16  20     27
    tokens = tokenizer.tokenize(data_source)

    assert (
        len(tokens) == 2
    ), "wrong number of tokens, expected: 2, found: {}".format(len(tokens))
    tok1, tok2 = tokens[0], tokens[1]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AaaaAaAaaAaAaaaa"
    ), "wrong data for token 1, expected: 'AaaaAaAaaAaAaaaa', found: {}".format(
        data
    )
    assert (
        start == 1
    ), "wrong start frame for token 1, expected: 1, found: {}".format(start)
    assert (
        end == 16
    ), "wrong end frame for token 1, expected: 16, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAAAAAA"
    ), "wrong data for token 2, expected: 'AAAAAAAA', found: {}".format(data)
    assert (
        start == 20
    ), "wrong start frame for token 2, expected: 20, found: {}".format(start)
    assert (
        end == 27
    ), "wrong end frame for token 2, expected: 27, found: {}".format(end)


def test_init_min_3_init_max_silence_0(validator):
    tokenizer = StreamTokenizer(
        validator,
        min_length=5,
        max_length=20,
        max_continuous_silence=4,
        init_min=3,
        init_max_silence=0,
        mode=0,
    )

    data_source = StringDataSource("aAaaaAaAaaAaAaaaaaAAAAAAAAAaaaaaaAAAAA")
    #                                                 ^           ^  ^   ^
    #                                                 18          30 33  37

    tokens = tokenizer.tokenize(data_source)

    assert (
        len(tokens) == 2
    ), "wrong number of tokens, expected: 2, found: {}".format(len(tokens))
    tok1, tok2 = tokens[0], tokens[1]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAAAAAAaaaa"
    ), "wrong data for token 1, expected: 'AAAAAAAAAaaaa', found: {}".format(
        data
    )
    assert (
        start == 18
    ), "wrong start frame for token 1, expected: 18, found: {}".format(start)
    assert (
        end == 30
    ), "wrong end frame for token 1, expected: 30, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAAA"
    ), "wrong data for token 2, expected: 'AAAAA', found: {}".format(data)
    assert (
        start == 33
    ), "wrong start frame for token 2, expected: 33, found: {}".format(start)
    assert (
        end == 37
    ), "wrong end frame for token 2, expected: 37, found: {}".format(end)


def test_init_min_3_init_max_silence_2(validator):
    tokenizer = StreamTokenizer(
        validator,
        min_length=5,
        max_length=20,
        max_continuous_silence=4,
        init_min=3,
        init_max_silence=2,
        mode=0,
    )

    data_source = StringDataSource("aAaaaAaAaaAaAaaaaaaAAAAAAAAAaaaaaaaAAAAA")
    #                                    ^          ^  ^           ^   ^   ^
    #                                    5          16 19          31  35  39
    tokens = tokenizer.tokenize(data_source)

    assert (
        len(tokens) == 3
    ), "wrong number of tokens, expected: 3, found: {}".format(len(tokens))
    tok1, tok2, tok3 = tokens[0], tokens[1], tokens[2]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AaAaaAaAaaaa"
    ), "wrong data for token 1, expected: 'AaAaaAaA', found: {}".format(data)
    assert (
        start == 5
    ), "wrong start frame for token 1, expected: 5, found: {}".format(start)
    assert (
        end == 16
    ), "wrong end frame for token 1, expected: 16, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAAAAAAAaaaa"
    ), "wrong data for token 2, expected: 'AAAAAAAAAaaaa', found: {}".format(
        data
    )
    assert (
        start == 19
    ), "wrong start frame for token 2, expected: 19, found: {}".format(start)
    assert (
        end == 31
    ), "wrong end frame for token 2, expected: 31, found: {}".format(end)

    data = "".join(tok3[0])
    start = tok3[1]
    end = tok3[2]
    assert (
        data == "AAAAA"
    ), "wrong data for token 3, expected: 'AAAAA', found: {}".format(data)
    assert (
        start == 35
    ), "wrong start frame for token 3, expected: 35, found: {}".format(start)
    assert (
        end == 39
    ), "wrong end frame for token 3, expected: 39, found: {}".format(end)


@pytest.fixture
def tokenizer_min_max_length(validator):
    return StreamTokenizer(
        validator,
        min_length=6,
        max_length=20,
        max_continuous_silence=2,
        init_min=3,
        init_max_silence=3,
        mode=0,
    )


def test_min_length_6_init_max_length_20(tokenizer_min_max_length):
    data_source = StringDataSource("aAaaaAaAaaAaAaaaaaAAAAAAAAAaaaaaAAAAA")
    #                                ^            ^   ^         ^
    #                                1            14  18        28

    tokens = tokenizer_min_max_length.tokenize(data_source)

    assert (
        len(tokens) == 2
    ), "wrong number of tokens, expected: 2, found: {}".format(len(tokens))
    tok1, tok2 = tokens[0], tokens[1]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AaaaAaAaaAaAaa"
    ), "wrong data for token 1, expected: 'AaaaAaAaaAaAaa', found: {}".format(
        data
    )
    assert (
        start == 1
    ), "wrong start frame for token 1, expected: 1, found: {}".format(start)
    assert (
        end == 14
    ), "wrong end frame for token 1, expected: 14, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAAAAAAAaa"
    ), "wrong data for token 2, expected: 'AAAAAAAAAaa', found: {}".format(data)
    assert (
        start == 18
    ), "wrong start frame for token 2, expected: 18, found: {}".format(start)
    assert (
        end == 28
    ), "wrong end frame for token 2, expected: 28, found: {}".format(end)


@pytest.fixture
def tokenizer_min_max_length_1_1(validator):
    return StreamTokenizer(
        validator,
        min_length=1,
        max_length=1,
        max_continuous_silence=0,
        init_min=0,
        init_max_silence=0,
        mode=0,
    )


def test_min_length_1_init_max_length_1(tokenizer_min_max_length_1_1):
    data_source = StringDataSource("AAaaaAaaaAaAaaAaAaaaaaAAAAAAAAAaaaaaAAAAA")

    tokens = tokenizer_min_max_length_1_1.tokenize(data_source)

    assert (
        len(tokens) == 21
    ), "wrong number of tokens, expected: 21, found: {}".format(len(tokens))


@pytest.fixture
def tokenizer_min_max_length_10_20(validator):
    return StreamTokenizer(
        validator,
        min_length=10,
        max_length=20,
        max_continuous_silence=4,
        init_min=3,
        init_max_silence=3,
        mode=0,
    )


def test_min_length_10_init_max_length_20(tokenizer_min_max_length_10_20):
    data_source = StringDataSource(
        "aAaaaAaAaaAaAaaaaaaAAAAAaaaaaaAAAAAaaAAaaAAA"
    )
    #     ^              ^             ^            ^
    #     1              16            30           45

    tokens = tokenizer_min_max_length_10_20.tokenize(data_source)

    assert (
        len(tokens) == 2
    ), "wrong number of tokens, expected: 2, found: {}".format(len(tokens))
    tok1, tok2 = tokens[0], tokens[1]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AaaaAaAaaAaAaaaa"
    ), "wrong data for token 1, expected: 'AaaaAaAaaAaAaaaa', found: {}".format(
        data
    )
    assert (
        start == 1
    ), "wrong start frame for token 1, expected: 1, found: {}".format(start)
    assert (
        end == 16
    ), "wrong end frame for token 1, expected: 16, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAAAaaAAaaAAA"
    ), "wrong data for token 2, expected: 'AAAAAaaAAaaAAA', found: {}".format(
        data
    )
    assert (
        start == 30
    ), "wrong start frame for token 2, expected: 30, found: {}".format(start)
    assert (
        end == 43
    ), "wrong end frame for token 2, expected: 43, found: {}".format(end)


@pytest.fixture
def tokenizer_min_max_length_4_5(validator):
    return StreamTokenizer(
        validator,
        min_length=4,
        max_length=5,
        max_continuous_silence=4,
        init_min=3,
        init_max_silence=3,
        mode=0,
    )


def test_min_length_4_init_max_length_5(tokenizer_min_max_length_4_5):
    data_source = StringDataSource(
        "aAaaaAaAaaAaAaaaaaAAAAAAAAaaaaaaAAAAAaaaaaAAaaAaa"
    )
    #                      ^   ^^   ^    ^   ^     ^   ^
    #                      18 2223  27   32  36    42  46

    tokens = tokenizer_min_max_length_4_5.tokenize(data_source)

    assert (
        len(tokens) == 4
    ), "wrong number of tokens, expected: 4, found: {}".format(len(tokens))
    tok1, tok2, tok3, tok4 = tokens[0], tokens[1], tokens[2], tokens[3]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAA"
    ), "wrong data for token 1, expected: 'AAAAA', found: {}".format(data)
    assert (
        start == 18
    ), "wrong start frame for token 1, expected: 18, found: {}".format(start)
    assert (
        end == 22
    ), "wrong end frame for token 1, expected: 22, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAaa"
    ), "wrong data for token 2, expected: 'AAAaa', found: {}".format(data)
    assert (
        start == 23
    ), "wrong start frame for token 2, expected: 23, found: {}".format(start)
    assert (
        end == 27
    ), "wrong end frame for token 2, expected: 27, found: {}".format(end)

    data = "".join(tok3[0])
    start = tok3[1]
    end = tok3[2]
    assert (
        data == "AAAAA"
    ), "wrong data for token 3, expected: 'AAAAA', found: {}".format(data)
    assert (
        start == 32
    ), "wrong start frame for token 3, expected: 32, found: {}".format(start)
    assert (
        end == 36
    ), "wrong end frame for token 3, expected: 36, found: {}".format(end)

    data = "".join(tok4[0])
    start = tok4[1]
    end = tok4[2]
    assert (
        data == "AAaaA"
    ), "wrong data for token 4, expected: 'AAaaA', found: {}".format(data)
    assert (
        start == 42
    ), "wrong start frame for token 4, expected: 42, found: {}".format(start)
    assert (
        end == 46
    ), "wrong end frame for token 4, expected: 46, found: {}".format(end)


@pytest.fixture
def tokenizer_max_continuous_silence_0(validator):
    return StreamTokenizer(
        validator,
        min_length=5,
        max_length=10,
        max_continuous_silence=0,
        init_min=3,
        init_max_silence=3,
        mode=0,
    )


def test_min_5_max_10_max_continuous_silence_0(
    tokenizer_max_continuous_silence_0,
):
    data_source = StringDataSource("aaaAAAAAaAAAAAAaaAAAAAAAAAa")
    #                                  ^   ^ ^    ^  ^       ^
    #                                  3   7 9   14 17      25

    tokens = tokenizer_max_continuous_silence_0.tokenize(data_source)

    assert (
        len(tokens) == 3
    ), "wrong number of tokens, expected: 3, found: {}".format(len(tokens))
    tok1, tok2, tok3 = tokens[0], tokens[1], tokens[2]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAA"
    ), "wrong data for token 1, expected: 'AAAAA', found: {}".format(data)
    assert (
        start == 3
    ), "wrong start frame for token 1, expected: 3, found: {}".format(start)
    assert (
        end == 7
    ), "wrong end frame for token 1, expected: 7, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAAAAA"
    ), "wrong data for token 2, expected: 'AAAAAA', found: {}".format(data)
    assert (
        start == 9
    ), "wrong start frame for token 2, expected: 9, found: {}".format(start)
    assert (
        end == 14
    ), "wrong end frame for token 2, expected: 14, found: {}".format(end)

    data = "".join(tok3[0])
    start = tok3[1]
    end = tok3[2]
    assert (
        data == "AAAAAAAAA"
    ), "wrong data for token 3, expected: 'AAAAAAAAA', found: {}".format(data)
    assert (
        start == 17
    ), "wrong start frame for token 3, expected: 17, found: {}".format(start)
    assert (
        end == 25
    ), "wrong end frame for token 3, expected: 25, found: {}".format(end)


@pytest.fixture
def tokenizer_max_continuous_silence_1(validator):
    return StreamTokenizer(
        validator,
        min_length=5,
        max_length=10,
        max_continuous_silence=1,
        init_min=3,
        init_max_silence=3,
        mode=0,
    )


def test_min_5_max_10_max_continuous_silence_1(
    tokenizer_max_continuous_silence_1,
):
    data_source = StringDataSource("aaaAAAAAaAAAAAAaaAAAAAAAAAa")
    #                                  ^        ^^ ^ ^        ^
    #                                  3       12131517      26
    #                                         (12 13 15 17)

    tokens = tokenizer_max_continuous_silence_1.tokenize(data_source)

    assert (
        len(tokens) == 3
    ), "wrong number of tokens, expected: 3, found: {}".format(len(tokens))
    tok1, tok2, tok3 = tokens[0], tokens[1], tokens[2]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAAaAAAA"
    ), "wrong data for token 1, expected: 'AAAAAaAAAA', found: {}".format(data)
    assert (
        start == 3
    ), "wrong start frame for token 1, expected: 3, found: {}".format(start)
    assert (
        end == 12
    ), "wrong end frame for token 1, expected: 12, found: {}".format(end)

    data = "".join(tok2[0])
    start = tok2[1]
    end = tok2[2]
    assert (
        data == "AAa"
    ), "wrong data for token 2, expected: 'AAa', found: {}".format(data)
    assert (
        start == 13
    ), "wrong start frame for token 2, expected: 13, found: {}".format(start)
    assert (
        end == 15
    ), "wrong end frame for token 2, expected: 15, found: {}".format(end)

    data = "".join(tok3[0])
    start = tok3[1]
    end = tok3[2]
    assert (
        data == "AAAAAAAAAa"
    ), "wrong data for token 3, expected: 'AAAAAAAAAa', found: {}".format(data)
    assert (
        start == 17
    ), "wrong start frame for token 3, expected: 17, found: {}".format(start)
    assert (
        end == 26
    ), "wrong end frame for token 3, expected: 26, found: {}".format(end)


@pytest.fixture
def tokenizer_strict_min_length(validator):
    return StreamTokenizer(
        validator,
        min_length=5,
        max_length=8,
        max_continuous_silence=3,
        init_min=3,
        init_max_silence=3,
        mode=StreamTokenizer.STRICT_MIN_LENGTH,
    )


def test_STRICT_MIN_LENGTH(tokenizer_strict_min_length):
    data_source = StringDataSource("aaAAAAAAAAAAAA")
    #                                 ^      ^
    #                                 2      9

    tokens = tokenizer_strict_min_length.tokenize(data_source)

    assert (
        len(tokens) == 1
    ), "wrong number of tokens, expected: 1, found: {}".format(len(tokens))
    tok1 = tokens[0]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAAAAA"
    ), "wrong data for token 1, expected: 'AAAAAAAA', found: {}".format(data)
    assert (
        start == 2
    ), "wrong start frame for token 1, expected: 2, found: {}".format(start)
    assert (
        end == 9
    ), "wrong end frame for token 1, expected: 9, found: {}".format(end)


@pytest.fixture
def tokenizer_drop_trailing_silence(validator):
    return StreamTokenizer(
        validator,
        min_length=5,
        max_length=10,
        max_continuous_silence=2,
        init_min=3,
        init_max_silence=3,
        mode=StreamTokenizer.DROP_TRAILING_SILENCE,
    )


def test_DROP_TAILING_SILENCE(tokenizer_drop_trailing_silence):
    data_source = StringDataSource("aaAAAAAaaaaa")
    #                                 ^   ^
    #                                 2   6

    tokens = tokenizer_drop_trailing_silence.tokenize(data_source)

    assert (
        len(tokens) == 1
    ), "wrong number of tokens, expected: 1, found: {}".format(len(tokens))
    tok1 = tokens[0]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAA"
    ), "wrong data for token 1, expected: 'AAAAA', found: {}".format(data)
    assert (
        start == 2
    ), "wrong start frame for token 1, expected: 2, found: {}".format(start)
    assert (
        end == 6
    ), "wrong end frame for token 1, expected: 6, found: {}".format(end)


@pytest.fixture
def tokenizer_strict_min_and_drop_trailing_silence(validator):
    return StreamTokenizer(
        validator,
        min_length=5,
        max_length=8,
        max_continuous_silence=3,
        init_min=3,
        init_max_silence=3,
        mode=StreamTokenizer.STRICT_MIN_LENGTH
        | StreamTokenizer.DROP_TRAILING_SILENCE,
    )


def test_STRICT_MIN_LENGTH_and_DROP_TAILING_SILENCE(
    tokenizer_strict_min_and_drop_trailing_silence,
):
    data_source = StringDataSource("aaAAAAAAAAAAAAaa")
    #                                 ^      ^
    #                                 2      8

    tokens = tokenizer_strict_min_and_drop_trailing_silence.tokenize(
        data_source
    )

    assert (
        len(tokens) == 1
    ), "wrong number of tokens, expected: 1, found: {}".format(len(tokens))
    tok1 = tokens[0]

    data = "".join(tok1[0])
    start = tok1[1]
    end = tok1[2]
    assert (
        data == "AAAAAAAA"
    ), "wrong data for token 1, expected: 'AAAAAAAA', found: {}".format(data)
    assert (
        start == 2
    ), "wrong start frame for token 1, expected: 2, found: {}".format(start)
    assert (
        end == 9
    ), "wrong end frame for token 1, expected: 9, found: {}".format(end)


@pytest.fixture
def tokenizer_callback(validator):
    return StreamTokenizer(
        validator,
        min_length=5,
        max_length=8,
        max_continuous_silence=3,
        init_min=3,
        init_max_silence=3,
        mode=0,
    )


def test_callback(tokenizer_callback):
    tokens = []

    def callback(data, start, end):
        tokens.append((data, start, end))

    data_source = StringDataSource("aaAAAAAAAAAAAAa")
    #                                 ^      ^^   ^
    #                                 2      910  14

    tokenizer_callback.tokenize(data_source, callback=callback)

    assert (
        len(tokens) == 2
    ), "wrong number of tokens, expected: 2, found: {}".format(len(tokens))


def test_max_length_inf_single_token(validator):
    """With max_length=inf, the entire valid sequence is one token."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=float("inf"),
        max_continuous_silence=3,
    )
    data_source = StringDataSource("aAAAaaAAAAAAAAAaaAAa")
    #                                ^                 ^
    #                                1                 18
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    data = "".join(tokens[0][0])
    assert data == "AAAaaAAAAAAAAAaaAAa"
    assert tokens[0][1] == 1
    assert tokens[0][2] == 19


def test_max_length_inf_splits_on_silence(validator):
    """With max_length=inf, tokens still split on max_continuous_silence."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=float("inf"),
        max_continuous_silence=2,
    )
    data_source = StringDataSource("AAAAAaaaAAAAA")
    #                               ^   ^   ^   ^
    #                               0   4   8   12
    # silence gap of 3 ('aaa') > max_continuous_silence (2) -> two tokens
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 2
    assert "".join(tokens[0][0]) == "AAAAAaa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 6
    assert "".join(tokens[1][0]) == "AAAAA"
    assert tokens[1][1] == 8
    assert tokens[1][2] == 12


def test_max_length_inf_respects_min_length(validator):
    """With max_length=inf, min_length is still enforced."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=5,
        max_length=float("inf"),
        max_continuous_silence=1,
        mode=StreamTokenizer.STRICT_MIN_LENGTH,
    )
    data_source = StringDataSource("AAaaAAAAAAAA")
    #                                   ^^^^^^^^
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAAAAA"
    assert tokens[0][1] == 4
    assert tokens[0][2] == 11


def test_max_length_inf_with_drop_trailing_silence(validator):
    """With max_length=inf and DROP_TRAILING_SILENCE, trailing silence
    is removed."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=float("inf"),
        max_continuous_silence=3,
        mode=StreamTokenizer.DROP_TRAILING_SILENCE,
    )
    data_source = StringDataSource("AAAAAaaa")
    #                               ^   ^
    #                               0   4
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAA"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 4


# ── max_leading_silence tests ──


def test_max_leading_silence_basic(validator):
    """Leading silence frames are prepended to the token."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        max_leading_silence=2,
    )
    data_source = StringDataSource("aaaAAAAAa")
    #                               012345678
    # buffer fills with frames 1,2 (maxlen=2), token starts at 1
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "aaAAAAAa"
    assert tokens[0][1] == 1
    assert tokens[0][2] == 8


def test_max_leading_silence_less_silence_than_requested(validator):
    """When fewer silent frames exist than max_leading_silence, use what's
    available."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        max_leading_silence=5,
    )
    data_source = StringDataSource("aAAAAAa")
    #                               0123456
    # only 1 silent frame available before first A
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "aAAAAAa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 6


def test_max_leading_silence_no_silence_before_token(validator):
    """Token at start of stream with no preceding silence."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        max_leading_silence=3,
    )
    data_source = StringDataSource("AAAAAaaa")
    #                               01234567
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAAa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 5


def test_max_leading_silence_two_tokens_no_overlap(validator):
    """Two tokens with leading silence; starts never overlap previous ends."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        max_leading_silence=2,
    )
    data_source = StringDataSource("aaAAAAAaaaAAAAAaa")
    #                               01234567890123456
    # Token 1: buffer=[a,a] from frames 0,1 → start=0
    # Token 2: buffer=[a] from frame 9 (1 frame in SILENCE after frame 8
    #          was consumed) → start=9

    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 2

    assert "".join(tokens[0][0]) == "aaAAAAAa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 7

    assert "".join(tokens[1][0]) == "aAAAAAa"
    assert tokens[1][1] == 9
    assert tokens[1][2] == 15

    # verify no overlap
    assert tokens[1][1] > tokens[0][2]


def test_max_leading_silence_wide_gap_between_tokens(validator):
    """With a wide silence gap, buffer captures the last N frames."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        max_leading_silence=2,
    )
    data_source = StringDataSource("aAAAaaaaaaAAAa")
    #                               01234567890123
    # Token 1: buffer=[a] from frame 0 → start=0
    # After token 1, SILENCE entered. buffer starts collecting.
    #   Frame 5 was consumed at the POSSIBLE_SILENCE→SILENCE transition.
    #   Frame 6: buffer=[a]
    #   Frame 7: buffer=[a,a]
    #   Frame 8: buffer=[a,a] (evicts 6, keeps 7,8)
    #   Frame 9: buffer=[a,a] (evicts 7, keeps 8,9)
    # Frame 10 (A): leading=[a,a] from frames 8,9 → start=10-2=8
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 2

    assert "".join(tokens[0][0]) == "aAAAa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 4

    assert "".join(tokens[1][0]) == "aaAAAa"
    assert tokens[1][1] == 8
    assert tokens[1][2] == 13

    assert tokens[1][1] > tokens[0][2]


def test_max_leading_silence_with_drop_trailing_silence(validator):
    """Leading silence is kept even when trailing silence is dropped."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=2,
        max_leading_silence=2,
        mode=StreamTokenizer.DROP_TRAILING_SILENCE,
    )
    data_source = StringDataSource("aaaAAAAAaaa")
    #                               01234567890
    # buffer=[a,a] from frames 1,2. Token starts at 1.
    # DROP_TRAILING removes the trailing 'aa' (silence_length=2 from
    # frames 8,9; frame 10 triggers SILENCE)
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "aaAAAAA"
    assert tokens[0][1] == 1
    assert tokens[0][2] == 7


def test_max_leading_silence_with_init_min_rejected_then_accepted(validator):
    """init_min rejects first attempt; leading silence is discarded and
    recollected for the successful detection."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        init_min=3,
        init_max_silence=0,
        max_leading_silence=2,
    )
    data_source = StringDataSource("aaAaaAAAAAa")
    #                               01234567890
    # Frame 0,1: SILENCE, buffer=[a,a]
    # Frame 2 (A): prepend buffer → data=[a,a,A], start=0. init_count=1 < 3
    #              → POSSIBLE_NOISE. buffer cleared.
    # Frame 3 (a): POSSIBLE_NOISE, sl=1 > init_max_silent=0 → SILENCE.
    #              data=[]. buffer is empty.
    # Frame 4 (a): SILENCE, buffer=[a]
    # Frame 5 (A): prepend buffer=[a] → data=[a,A], start=4. init_count=1 < 3
    #              → POSSIBLE_NOISE. buffer cleared.
    # Frame 6 (A): POSSIBLE_NOISE, valid. init_count=2.
    # Frame 7 (A): POSSIBLE_NOISE, valid. init_count=3 >= 3 → NOISE.
    # Frame 8 (A): NOISE.
    # Frame 9 (A): NOISE. data=[a,A,A,A,A,A]
    # Frame 10(a): POSSIBLE_SILENCE. sl=1>=mcs(1) would need another frame...
    #   Actually mcs=1, so sl=1 is OK. data=[a,A,A,A,A,A,a]
    #   Wait, this needs a second silent frame to trigger SILENCE.
    #   End of data → _post_process → deliver.
    # Token: data=[a,A,A,A,A,A,a], start=4, end=4+7-1=10
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "aAAAAAa"
    assert tokens[0][1] == 4
    assert tokens[0][2] == 10


def test_max_leading_silence_with_init_min_accepted(validator):
    """init_min and max_leading_silence compose: leading silence is kept
    when init_min is satisfied."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=1,
        init_min=3,
        init_max_silence=1,
        max_leading_silence=2,
    )
    data_source = StringDataSource("aaaAaAAAa")
    #                               012345678
    # Frame 0,1: SILENCE, buffer=[a,a]
    # Frame 2 (a): SILENCE, buffer=[a,a] (evicts 0, keeps 1,2)
    # Frame 3 (A): prepend [a,a] → data=[a,a,A], start=1. init=1 < 3
    #              → POSSIBLE_NOISE. buffer cleared.
    # Frame 4 (a): POSSIBLE_NOISE, sl=1 <= init_max_silent(1). data=[a,a,A,a]
    # Frame 5 (A): POSSIBLE_NOISE, valid. sl=0. init=2 < 3. data=[a,a,A,a,A]
    # Frame 6 (A): POSSIBLE_NOISE, valid. init=3 >= 3 → NOISE. data=[a,a,A,a,A,A]
    # Frame 7 (A): NOISE. data=[a,a,A,a,A,A,A]
    # Frame 8 (a): POSSIBLE_SILENCE. sl=1>=mcs(1)... no wait, sl=1 and mcs=1.
    #   Actually: sl is set to 1. 1 >= 1 is checked on the NEXT silent frame.
    #   End of data → _post_process. state=POSSIBLE_SILENCE, len(data)=8 > sl=1.
    #   → deliver.
    # Token: data=[a,a,A,a,A,A,A,a], start=1, end=1+8-1=8
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "aaAaAAAa"
    assert tokens[0][1] == 1
    assert tokens[0][2] == 8


def test_max_leading_silence_0_is_backward_compatible(validator):
    """max_leading_silence=0 produces identical results to omitting it."""
    tokenizer_default = StreamTokenizer(
        validator,
        min_length=5,
        max_length=20,
        max_continuous_silence=4,
        init_min=3,
        init_max_silence=2,
    )
    tokenizer_explicit = StreamTokenizer(
        validator,
        min_length=5,
        max_length=20,
        max_continuous_silence=4,
        init_min=3,
        init_max_silence=2,
        max_leading_silence=0,
    )
    data = "aAaaaAaAaaAaAaaaaaaAAAAAAAAAaaaaaaaAAAAA"
    tokens_default = tokenizer_default.tokenize(StringDataSource(data))
    tokens_explicit = tokenizer_explicit.tokenize(StringDataSource(data))
    assert len(tokens_default) == len(tokens_explicit)
    for t1, t2 in zip(tokens_default, tokens_explicit):
        assert "".join(t1[0]) == "".join(t2[0])
        assert t1[1] == t2[1]
        assert t1[2] == t2[2]


# ── max_trailing_silence tests ──


def test_max_trailing_silence_none_keeps_all(validator):
    """max_trailing_silence=None (default) keeps all trailing silence."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=3,
    )
    data_source = StringDataSource("AAAAAaaaa")
    #                               012345678
    # mcs=3: frames 5,6,7 tolerated, frame 8 triggers SILENCE
    # Token includes trailing aaa (3 frames)
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAAaaa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 7


def test_max_trailing_silence_0_drops_all(validator):
    """max_trailing_silence=0 drops all trailing silence."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=3,
        max_trailing_silence=0,
    )
    data_source = StringDataSource("AAAAAaaaa")
    #                               012345678
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAA"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 4


def test_max_trailing_silence_partial(validator):
    """max_trailing_silence=1 keeps only 1 frame of trailing silence."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=3,
        max_trailing_silence=1,
    )
    data_source = StringDataSource("AAAAAaaaa")
    #                               012345678
    # 3 trailing frames, keep only 1 → "AAAAAa"
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAAa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 5


def test_max_trailing_silence_larger_than_actual(validator):
    """When max_trailing_silence > actual trailing silence, keep all."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=2,
        max_trailing_silence=5,
    )
    data_source = StringDataSource("AAAAAaaa")
    #                               01234567
    # mcs=2: frames 5,6 tolerated, frame 7 triggers SILENCE
    # 2 trailing frames < max_trailing_silence(5), keep all
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "AAAAAaa"
    assert tokens[0][1] == 0
    assert tokens[0][2] == 6


def test_max_trailing_silence_0_matches_drop_trailing_mode(validator):
    """max_trailing_silence=0 gives the same result as DROP_TRAILING_SILENCE
    mode flag."""
    tokenizer_mode = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=3,
        mode=StreamTokenizer.DROP_TRAILING_SILENCE,
    )
    tokenizer_param = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=3,
        max_trailing_silence=0,
    )
    data = "aaAAAAAaaaAAAa"
    tokens_mode = tokenizer_mode.tokenize(StringDataSource(data))
    tokens_param = tokenizer_param.tokenize(StringDataSource(data))
    assert len(tokens_mode) == len(tokens_param)
    for t1, t2 in zip(tokens_mode, tokens_param):
        assert "".join(t1[0]) == "".join(t2[0])
        assert t1[1] == t2[1]
        assert t1[2] == t2[2]


def test_max_trailing_silence_with_max_leading_silence(validator):
    """Both leading and trailing silence control work together."""
    tokenizer = StreamTokenizer(
        validator,
        min_length=1,
        max_length=20,
        max_continuous_silence=3,
        max_leading_silence=2,
        max_trailing_silence=1,
    )
    data_source = StringDataSource("aaaAAAAAaaaa")
    #                               0123456789 10 11
    # buffer=[a,a] from frames 1,2. Leading prepended → start=1
    # mcs=3: frames 8,9,10 tolerated, frame 11 triggers SILENCE
    # 3 trailing frames, keep 1 → trim 2
    tokens = tokenizer.tokenize(data_source)
    assert len(tokens) == 1
    assert "".join(tokens[0][0]) == "aaAAAAAa"
    assert tokens[0][1] == 1
    assert tokens[0][2] == 8



"""Integration tests for CLI subcommand dispatch, trim, and fix-pauses."""

import os
from tempfile import TemporaryDirectory

import pytest

from auditok.cmdline import main

WAV_FILE = "tests/data/test_16KHZ_mono_400Hz.wav"


# ── Subcommand dispatch ────────────────────────────────────────────


class TestSubcommandDispatch:
    def test_version(self, capsys):
        ret = main(["--version"])
        assert ret == 0
        assert "0.4.0" in capsys.readouterr().out

    def test_version_short(self, capsys):
        ret = main(["-v"])
        assert ret == 0
        assert "0.4.0" in capsys.readouterr().out

    def test_help_shows_subcommands(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            main(["--help"])
        assert exc_info.value.code == 0
        out = capsys.readouterr().out
        assert "split" in out
        assert "trim" in out
        assert "fix-pauses" in out

    def test_split_help(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            main(["split", "--help"])
        assert exc_info.value.code == 0
        out = capsys.readouterr().out
        assert "auditok split" in out

    def test_trim_help(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            main(["trim", "--help"])
        assert exc_info.value.code == 0
        out = capsys.readouterr().out
        assert "auditok trim" in out

    def test_fix_pauses_help(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            main(["fix-pauses", "--help"])
        assert exc_info.value.code == 0
        out = capsys.readouterr().out
        assert "auditok fix-pauses" in out

    def test_implicit_split_with_file(self, capsys):
        """auditok file.wav ... should behave like auditok split file.wav ..."""
        ret = main([WAV_FILE, "-q"])
        assert ret == 0

    def test_explicit_split_with_file(self, capsys):
        ret = main(["split", WAV_FILE, "-q"])
        assert ret == 0


# ── trim subcommand ────────────────────────────────────────────────


class TestTrimSubcommand:
    def test_trim_file_creates_output(self):
        with TemporaryDirectory() as tmpdir:
            output = os.path.join(tmpdir, "trimmed.wav")
            ret = main(["trim", WAV_FILE, "-o", output])
            assert ret == 0
            assert os.path.exists(output)
            assert os.path.getsize(output) > 0

    def test_trim_no_activity(self, capsys):
        with TemporaryDirectory() as tmpdir:
            output = os.path.join(tmpdir, "trimmed.wav")
            ret = main(["trim", WAV_FILE, "-o", output, "-e", "99"])
            assert ret == 0
            captured = capsys.readouterr()
            assert "No audio activity" in captured.err

    def test_trim_with_output_format(self):
        with TemporaryDirectory() as tmpdir:
            output = os.path.join(tmpdir, "trimmed.raw")
            ret = main(["trim", WAV_FILE, "-o", output, "-T", "raw"])
            assert ret == 0
            assert os.path.exists(output)
            assert os.path.getsize(output) > 0

    def test_trim_requires_output(self):
        with pytest.raises(SystemExit) as exc_info:
            main(["trim", WAV_FILE])
        assert exc_info.value.code != 0


# ── fix-pauses subcommand ──────────────────────────────────────────


class TestFixPausesSubcommand:
    def test_fix_pauses_file_creates_output(self):
        with TemporaryDirectory() as tmpdir:
            output = os.path.join(tmpdir, "fixed.wav")
            ret = main(["fix-pauses", WAV_FILE, "-o", output, "-d", "0.5"])
            assert ret == 0
            assert os.path.exists(output)
            assert os.path.getsize(output) > 0

    def test_fix_pauses_no_activity(self, capsys):
        with TemporaryDirectory() as tmpdir:
            output = os.path.join(tmpdir, "fixed.wav")
            ret = main(
                ["fix-pauses", WAV_FILE, "-o", output, "-d", "0.5", "-e", "99"]
            )
            assert ret == 0
            captured = capsys.readouterr()
            assert "No audio activity" in captured.err

    def test_fix_pauses_requires_output(self):
        with pytest.raises(SystemExit) as exc_info:
            main(["fix-pauses", WAV_FILE, "-d", "0.5"])
        assert exc_info.value.code != 0

    def test_fix_pauses_requires_pause_duration(self):
        with pytest.raises(SystemExit) as exc_info:
            main(["fix-pauses", WAV_FILE, "-o", "out.wav"])
        assert exc_info.value.code != 0


# ── Deprecation warnings ──────────────────────────────────────────


class TestDeprecationWarnings:
    def test_join_detections_warns(self, capsys):
        with TemporaryDirectory() as tmpdir:
            stream = os.path.join(tmpdir, "stream.wav")
            ret = main(["split", WAV_FILE, "-q", "-j", "0.5", "-O", stream])
            assert ret == 0
            captured = capsys.readouterr()
            assert "--join-detections" in captured.err
            assert "deprecated" in captured.err.lower()



"""
Module for high-level audio input-output operations.

.. autosummary::
    :toctree: generated/

    AudioEnergyValidator
    AudioReader
    Recorder
    make_duration_formatter
    make_channel_selector
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import partial
from typing import Any, Callable

import numpy as np

from . import signal
from .exceptions import TimeFormatError, TooSmallBlockDuration
from .io import AudioIOError, AudioSource, BufferAudioSource, get_audio_source

__all__ = [
    "make_duration_formatter",
    "make_channel_selector",
    "DataSource",
    "DataValidator",
    "StringDataSource",
    "AudioReader",
    "Recorder",
    "AudioEnergyValidator",
]


def make_duration_formatter(fmt: str) -> Callable[[float], str]:
    """
    Create and return a function to format durations in seconds using a
    specified format. Accepted format directives are:

    - ``%S`` : absolute seconds with 3 decimals; must be used alone.
    - ``%i`` : milliseconds
    - ``%s`` : seconds
    - ``%m`` : minutes
    - ``%h`` : hours

    The last four directives (%i, %s, %m, %h) should all be specified and can
    be placed in any order within the input format string.

    Parameters
    ----------
    fmt : str
        Format string specifying the duration format.

    Returns
    -------
    formatter : callable
        A function that takes a duration in seconds (float) and returns a
        formatted string.

    Raises
    ------
    TimeFormatError
        Raised if the format contains an unknown directive.

    Examples
    --------
    Using ``%S`` for total seconds with three decimal precision:

    .. code:: python

        formatter = make_duration_formatter("%S")
        formatter(123.589)
        # '123.589'
        formatter(123)
        # '123.000'

    Using combined directives:

    .. code:: python

        formatter = make_duration_formatter("%h:%m:%s.%i")
        formatter(3723.25)
        # '01:02:03.250'

        formatter = make_duration_formatter("%h hrs, %m min, %s sec and %i ms")
        formatter(3723.25)
        # '01 hrs, 02 min, 03 sec and 250 ms'

    Note:
        Omitting any of the four main directives (%i, %s, %m, %h) may result
        in incorrect formatting:

        .. code:: python

            formatter = make_duration_formatter("%m min, %s sec and %i ms")
            formatter(3723.25)
            # '02 min, 03 sec and 250 ms'
    """

    if fmt == "%S":

        def formatter(seconds):
            return "{:.3f}".format(seconds)

    elif fmt == "%I":

        def formatter(seconds):
            return "{0}".format(int(seconds * 1000))

    else:
        fmt = fmt.replace("%h", "{hrs:02d}")
        fmt = fmt.replace("%m", "{mins:02d}")
        fmt = fmt.replace("%s", "{secs:02d}")
        fmt = fmt.replace("%i", "{millis:03d}")
        try:
            i = fmt.index("%")
            raise TimeFormatError(
                "Unknown time format directive '{0}'".format(fmt[i : i + 2])
            )
        except ValueError:
            pass

        def formatter(seconds):
            millis = int(seconds * 1000)
            hrs, millis = divmod(millis, 3600000)
            mins, millis = divmod(millis, 60000)
            secs, millis = divmod(millis, 1000)
            return fmt.format(hrs=hrs, mins=mins, secs=secs, millis=millis)

    return formatter


def make_channel_selector(
    sample_width: int, channels: int, selected: int | str | None = None
) -> Callable[[bytes], Any]:
    """
    Create and return a callable for selecting a specific audio channel. The
    returned `selector` function can be used as `selector(audio_data)` and
    returns data for the specified channel.

    If `selected` is None or "any", the `selector` will separate and return a
    list of available channels: `[data_channel_1, data_channel_2, ...]`.

    Note that `selector` expects input data in `bytes` format but does not
    necessarily return a `bytes` object. To select or compute the desired
    channel (or average channel if `selected="avg"`), it converts the input
    data into an `array.array` or `numpy.ndarray`. After selection, the data
    is returned as is, without reconversion to `bytes`, for efficiency. The
    output can be converted back to `bytes` with `bytes(obj)` if needed.

    Special case: If `channels=1`, the input data is returned without processing.

    Parameters
    ----------
    sample_width : int
        Number of bytes per audio sample; should be 1, 2, or 4.
    channels : int
        Number of channels in the audio data that the selector should expect.
    selected : int or str, optional
        Channel to select in each call to `selector(raw_data)`. Acceptable values:
        - An integer in range [-channels, channels).
        - "mix", "avg", or "average" for averaging across channels.
        - None or "any" to return a list of all channels.

    Returns
    -------
    selector : callable
        A function that can be called as `selector(audio_data)` and returns data
        for the selected channel.

    Raises
    ------
    ValueError
        If `sample_width` is not one of {1, 2, 4}, or if `selected` has an
        unsupported value.
    """
    to_array_ = partial(
        signal.to_array, sample_width=sample_width, channels=channels
    )
    if channels == 1 or selected in (None, "any"):
        return to_array_

    if isinstance(selected, int):
        if selected < 0:
            selected += channels
        if selected < 0 or selected >= channels:
            err_msg = "Selected channel must be >= -channels and < channels"
            err_msg += ", given: {}"
            raise ValueError(err_msg.format(selected))
        return lambda x: to_array_(x)[selected]

    if selected in ("mix", "avg", "average"):
        return lambda x: to_array_(x).mean(axis=0)

    raise ValueError(
        "Selected channel must be an integer, None (alias 'any') or 'average' "
        "(alias 'avg' or 'mix')"
    )


class DataSource(ABC):
    """
    Base class for objects used as data sources in
    :func:`StreamTokenizer.tokenize`.

    Subclasses should implement a :func:`DataSource.read` method, which is
    expected to return a frame (or slice) of data from the source, and None
    when there is no more data to read.
    """

    @abstractmethod
    def read(self) -> Any:
        """
        Read a block (or window) of data from this source.

        Returns
        -------
        data : object or None
            A block of data from the source. If no more data is available,
            should return None.
        """


class DataValidator(ABC):
    """
    Base class for validator objects used by :class:`.core.StreamTokenizer`
    to verify the validity of read data.

    Subclasses should implement the :func:`is_valid` method to define the
    specific criteria for data validity.
    """

    @abstractmethod
    def is_valid(self, data: Any) -> bool:
        """
        Determine whether the provided `data` meets validity criteria.

        Parameters
        ----------
        data : object
            The data to be validated.

        Returns
        -------
        bool
            True if `data` is valid, otherwise False.
        """


class AudioEnergyValidator(DataValidator):
    """
    A validator based on audio signal energy. For an input window of `N` audio
    samples (see :func:`AudioEnergyValidator.is_valid`), the energy is computed
    as:

    .. math::
        \\text{energy} = 20 \\log(\\sqrt({1}/{N} \\sum_{i=1}^{N} {a_i}^2))  % # noqa: W605

    where `a_i` represents the i-th audio sample.

    Parameters
    ----------
    energy_threshold : float
        Minimum energy required for an audio window to be considered valid.
    sample_width : int
        Size in bytes of a single audio sample.
    channels : int
        Number of audio channels in the data.
    use_channel : {None, "any", "mix", "avg", "average"} or int
        Specifies the channel used for energy computation:

        - None or "any": Compute energy for each channel and return the maximum.
        - "mix" (or "avg" / "average"): Average across all channels, then
          compute energy.
        - int (0 <= value < `channels`): Compute energy for the specified channel
          only, ignoring others.

    Returns
    -------
    energy : float
        Computed energy of the audio window, used to validate if the window
        meets the `energy_threshold`.
    """

    def __init__(
        self,
        energy_threshold: float,
        sample_width: int,
        channels: int,
        use_channel: int | str | None = None,
    ) -> None:
        self._energy_threshold = energy_threshold
        self._sample_width = sample_width
        self._selector = make_channel_selector(
            sample_width, channels, use_channel
        )
        self._energy_agg_fn = np.max if use_channel in (None, "any") else None

    def is_valid(self, data: bytes) -> bool:
        """
        Determine if the audio data meets the energy threshold.

        Parameters
        ----------
        data : bytes-like
            An array of raw audio data.

        Returns
        -------
        bool
            True if the energy of the audio data is greater than or equal to
            the specified threshold; otherwise, False.
        """

        log_energy = signal.calculate_energy(
            self._selector(data), self._energy_agg_fn
        )
        return log_energy >= self._energy_threshold


class StringDataSource(DataSource):
    """
    A :class:`DataSource` implementation that reads from a string buffer.

    Each call to :meth:`read` returns one character from the buffer and advances
    by one position. When the end of the buffer is reached, :meth:`read` returns
    None.

    Parameters
    ----------
    data : str
        The string data to be used as the source.
    """

    def __init__(self, data: str) -> None:

        self._data: str = ""
        self._current = 0
        self.set_data(data)

    def read(self) -> str | None:
        """
        Read one character from buffer.

        Returns
        -------
        char : str
            current character or None if the end of the buffer is reached.
        """

        if self._current >= len(self._data):
            return None
        self._current += 1
        return self._data[self._current - 1]

    def set_data(self, data: str) -> None:
        """
        Set a new data buffer.

        Parameters
        ----------
        data : str
            new data buffer.
        """

        if not isinstance(data, str):
            raise ValueError("data must an instance of str")
        self._data = data
        self._current = 0


class _AudioReadingProxy:
    def __init__(self, audio_source):

        self._audio_source = audio_source

    def rewind(self):
        if self.rewindable:
            self._audio_source.rewind()
        else:
            raise AudioIOError("Audio stream is not rewindable")

    def rewindable(self):
        try:
            return self._audio_source.rewindable
        except AttributeError:
            return False

    def is_open(self):
        return self._audio_source.is_open()

    def open(self):
        self._audio_source.open()

    def close(self):
        self._audio_source.close()

    def read(self, size):
        return self._audio_source.read(size)

    @property
    def data(self):
        err_msg = "This AudioReader is not a recorder, no recorded data can "
        err_msg += "be retrieved"
        raise AttributeError(err_msg)

    def __getattr__(self, name):
        return getattr(self._audio_source, name)


class _Recorder(_AudioReadingProxy):
    """
    A class for `AudioReader` objects that records all data read from the source.

    This class is particularly useful for capturing audio data when reading from
    a microphone or similar live audio sources.
    """

    def __init__(self, audio_source):
        super().__init__(audio_source)
        self._cache = []
        self._read_block = self._read_and_cache
        self._read_from_cache = False
        self._data = None

    def read(self, size):
        return self._read_block(size)

    @property
    def data(self):
        if self._data is None:
            err_msg = "Un-rewinded recorder. `rewind` should be called before "
            err_msg += "accessing recorded data"
            raise RuntimeError(err_msg)
        return self._data

    def rewindable(self):
        return True

    def rewind(self):
        if self._read_from_cache:
            self._audio_source.rewind()
        else:
            self._data = b"".join(self._cache)
            self._cache = None
            self._audio_source = BufferAudioSource(
                self._data, self.sr, self.sw, self.ch
            )
            self._read_block = self._audio_source.read
            self.open()
            self._read_from_cache = True

    def _read_and_cache(self, size):
        # Read and save read data
        block = self._audio_source.read(size)
        if block is not None:
            self._cache.append(block)
        return block


class _Limiter(_AudioReadingProxy):
    """
    A class for `AudioReader` objects that restricts the amount of data read.

    This class is useful for limiting data intake when reading from a microphone
    or large audio files, ensuring only a specified amount of data is processed.
    """

    def __init__(self, audio_source, max_read):
        super().__init__(audio_source)
        self._max_read = max_read
        self._max_samples = round(max_read * self.sr)
        self._bytes_per_sample = self.sw * self.ch
        self._read_samples = 0

    @property
    def data(self):
        data = self._audio_source.data
        max_read_bytes = self._max_samples * self._bytes_per_sample
        return data[:max_read_bytes]

    @property
    def max_read(self):
        return self._max_read

    def read(self, size):
        size = min(self._max_samples - self._read_samples, size)
        if size <= 0:
            return None
        block = self._audio_source.read(size)
        if block is None:
            return None
        self._read_samples += len(block) // self._bytes_per_sample
        return block

    def rewind(self):
        super().rewind()
        self._read_samples = 0


class _FixedSizeAudioReader(_AudioReadingProxy):
    """
    A class to read fixed-size audio windows from a source.
    """

    def __init__(self, audio_source, block_dur):
        super().__init__(audio_source)

        if block_dur <= 0:
            raise ValueError(
                "block_dur must be > 0, given: {}".format(block_dur)
            )

        self._block_size = int(block_dur * self.sr)
        if self._block_size == 0:
            err_msg = "Too small block_dur ({0:f}) for sampling rate ({1}). "
            err_msg += "block_dur should cover at least one sample "
            err_msg += "(i.e. 1/{1})"
            raise TooSmallBlockDuration(
                err_msg.format(block_dur, self.sr), block_dur, self.sr
            )

    def read(self):
        return self._audio_source.read(self._block_size)

    @property
    def block_size(self):
        return self._block_size

    @property
    def block_dur(self):
        return self._block_size / self.sr

    def __getattr__(self, name):
        return getattr(self._audio_source, name)


class _OverlapAudioReader(_FixedSizeAudioReader):
    """
    A class for `AudioReader` objects that reads and returns overlapping audio
    windows.

    Useful for applications requiring overlapping segments, such as audio
    analysis or feature extraction.
    """

    def __init__(self, audio_source, block_dur, hop_dur):

        if hop_dur >= block_dur:
            raise ValueError('"hop_dur" should be <= "block_dur"')

        super().__init__(audio_source, block_dur)

        self._hop_size = int(hop_dur * self.sr)
        if self._hop_size < 1:
            err_msg = (
                "'hop_dur' ({}) is too small for sampling rate ({}): "
                "int({} * {}) = {} (must be >= 1 sample). "
                "Use a larger 'hop_dur' or a higher sampling rate."
            ).format(hop_dur, self.sr, hop_dur, self.sr, self._hop_size)
            raise ValueError(err_msg)
        if self._hop_size >= self._block_size:
            err_msg = (
                "hop_size (int({} * {}) = {}) >= block_size "
                "(int({} * {}) = {}) for sampling rate ({}). "
                "Overlap requires hop_size < block_size. "
                "Increase the gap between 'hop_dur' and 'block_dur' "
                "or use a higher sampling rate."
            ).format(
                hop_dur,
                self.sr,
                self._hop_size,
                block_dur,
                self.sr,
                self._block_size,
                self.sr,
            )
            raise ValueError(err_msg)
        self._blocks = self._iter_blocks_with_overlap()

    def _iter_blocks_with_overlap(self):
        while not self.is_open():
            yield AudioIOError
        block = self._audio_source.read(self._block_size)
        if block is None:
            yield None

        _hop_size_bytes = (
            self._hop_size * self._audio_source.sw * self._audio_source.ch
        )
        cache = block[_hop_size_bytes:]
        yield block

        while True:
            block = self._audio_source.read(self._hop_size)
            if block:
                block = cache + block
                cache = block[_hop_size_bytes:]
                yield block
                continue
            yield None

    def read(self):
        try:
            block = next(self._blocks)
            if block == AudioIOError:
                raise AudioIOError("Audio Stream is not open.")
            return block
        except StopIteration:
            return None

    def rewind(self):
        super().rewind()
        self._blocks = self._iter_blocks_with_overlap()

    @property
    def hop_size(self):
        return self._hop_size

    @property
    def hop_dur(self):
        return self._hop_size / self.sr

    def __getattr__(self, name):
        return getattr(self._audio_source, name)


class AudioReader(DataSource):
    """
    A class to read fixed-size chunks of audio data from a source, which can
    be a file, standard input (with `input` set to "-"), or a microphone.
    Typically used by tokenization algorithms that require source objects with
    a `read` function to return data windows of consistent size, except for
    the last window if remaining data is insufficient.

    This class supports overlapping audio windows, recording the audio stream
    for later access (useful for microphone input), and limiting the maximum
    amount of data read.

    Parameters
    ----------
    input : str, bytes, AudioSource, AudioReader, AudioRegion, or None
        Input audio data. If a string, it should be the path to an audio file
        (use "-" for standard input). If bytes, the input is treated as raw
        audio data. If None, audio is read from a microphone. Any input that
        is not an :class:`AudioReader` will be converted, if possible, to an
        :class:`AudioSource` for processing. For raw audio (string path, bytes,
        or None), specify audio parameters using kwargs (`sampling_rate`,
        `sample_width`, `channels` or their aliases: `sr`, `sw`, `ch`).
    block_dur : float, default=0.01
        Duration of audio data (in seconds) to return in each `read` call.
    hop_dur : float, optional
        Duration of data to skip (in seconds) from the previous window. If set,
        it is used to calculate temporal overlap between the current and
        previous window (`overlap = block_dur - hop_dur`). If None (default),
        windows do not overlap.
    record : bool, default=False
        Whether to record audio data for later access. If True, recorded audio
        can be accessed using the `data` property after calling `rewind()`.
        Note: after `rewind()`, no new data is read from the source—subsequent
        `read` calls use the cached data.
    max_read : float, optional
        Maximum duration of audio data to read (in seconds). If None (default),
        data is read until the end of the stream or, for microphone input, until
        a Ctrl-C interruption.

    Additional audio parameters may be required if `input` is raw audio
    (None, bytes, or raw audio file):

    Other Parameters
    ----------------
    audio_format, fmt : str
        Type of audio data (e.g., wav, ogg, flac, raw). Used if `input` is a
        file path. If not provided, the format is inferred from the file
        extension or header.
    sampling_rate, sr : int
        Sampling rate of the audio data. Required for raw audio (bytes, None,
        or raw file).
    sample_width, sw : int
        Number of bytes per audio sample (typically 1, 2, or 4). Required for
        raw data.
    channels, ch : int
        Number of audio channels. Required for raw data.
    use_channel, uc : {None, "any", "mix", "avg", "average"} or int
        Specifies the channel used for split if `input` has multiple channels.
        All returned audio data includes data from *all* input channels. Options:

        - None or "any": Use any active channel, regardless of silence in others.
          (Default)
        - "mix" / "avg" / "average": Combine all channels by averaging.
        - int: Use the specified channel ID (0 <= value < `channels`).

    large_file : bool, default=False
        If True and `input` is a path to a *wav* or *raw* file, audio data is
        loaded lazily (one analysis window at a time). Otherwise, the entire
        file is loaded before processing. Use True for large files exceeding
        available memory.
    """

    _audio_source: Any

    def __init__(
        self,
        input: Any,
        block_dur: float = 0.01,
        hop_dur: float | None = None,
        record: bool = False,
        max_read: float | None = None,
        **kwargs: Any,
    ) -> None:
        if isinstance(input, AudioReader):
            # Unwrap to the base AudioSource so the proxy chain
            # (_Recorder, _Limiter, _FixedSizeAudioReader) is rebuilt
            # cleanly with the new parameters.
            source: Any = input._audio_source
            while isinstance(source, _AudioReadingProxy):
                source = source._audio_source
            input = source
        if not isinstance(input, AudioSource):
            input = get_audio_source(input, **kwargs)
        self._record = record
        if record:
            input = _Recorder(input)
        if max_read is not None:
            input = _Limiter(input, max_read)
            self._max_read = max_read
        if hop_dur is None or hop_dur == block_dur:
            input = _FixedSizeAudioReader(input, block_dur)
        else:
            input = _OverlapAudioReader(input, block_dur, hop_dur)

        self._audio_source = input

    def __repr__(self) -> str:
        block_dur, hop_dur, max_read = None, None, None
        if self.block_dur is not None:
            block_dur = "{:.3f}".format(self.block_dur)
        if self.hop_dur is not None:
            hop_dur = "{:.3f}".format(self.hop_dur)
        if self.max_read is not None:
            max_read = "{:.3f}".format(self.max_read)
        return (
            "<{cls}(block_dur={block_dur}, "
            "hop_dur={hop_dur}, record={rewindable}, "
            "max_read={max_read})>"
        ).format(
            cls=self.__class__.__name__,
            block_dur=block_dur,
            hop_dur=hop_dur,
            rewindable=self._record,
            max_read=max_read,
        )

    @property
    def rewindable(self) -> bool:
        return self._record

    @property
    def block_dur(self) -> float:
        return self._audio_source.block_size / self._audio_source.sr

    @property
    def hop_dur(self) -> float:
        if hasattr(self._audio_source, "hop_dur"):
            return self._audio_source.hop_size / self._audio_source.sr
        return self.block_dur

    @property
    def hop_size(self) -> int:
        if hasattr(self._audio_source, "hop_size"):
            return self._audio_source.hop_size
        return self.block_size

    @property
    def max_read(self) -> float | None:
        try:
            return self._audio_source.max_read
        except AttributeError:
            return None

    def read(self) -> bytes | None:
        return self._audio_source.read()  # type: ignore[call-arg]

    def __getattr__(self, name):
        if name in ("data", "rewind") and not self.rewindable:
            raise AttributeError(
                "'AudioReader' has no attribute '{}'".format(name)
            )
        try:
            return getattr(self._audio_source, name)
        except AttributeError as exc:
            raise AttributeError(
                f"'AudioReader' has no attribute {name!r}"
            ) from exc


class Recorder(AudioReader):
    """
    A class to read fixed-size chunks of audio data from a source and store
    them in a cache. This class is equivalent to initializing
    :class:`AudioReader` with `record=True`. For more details on additional
    parameters, refer to :class:`AudioReader`.

    Once the desired amount of data is read, you can call the :meth:`rewind`
    method to access the recorded data via the :attr:`data` attribute. The
    cached data can also be re-read in fixed-size windows by calling
    :meth:`read`.
    """

    def __init__(
        self,
        input: Any,
        block_dur: float = 0.01,
        hop_dur: float | None = None,
        max_read: float | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            input,
            block_dur=block_dur,
            hop_dur=hop_dur,
            record=True,
            max_read=max_read,
            **kwargs,
        )
